

class PromiumException(Exception):
    pass


class PromiumTimeout(PromiumException):
    def __init__(self, message="no error message", seconds=10, screen=None):
        self.message = f"{message} (waited {seconds} seconds)"
        if screen:
            self.message += "Screenshot: available via screen\n"
        super(PromiumTimeout, self).__init__(self.message)


class ElementLocationException(PromiumException):
    pass


class LocatorException(PromiumException):
    pass


class BrowserConsoleException(PromiumException):
    pass




import requests

from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from promium.base import Element
from promium.helpers import ParseUrl


class Link(Element):

    @property
    def href(self):
        return self.get_attribute("href")

    @property
    def parse_url(self):
        """
        Using parse_url to parse href. Can get attributes:
        - scheme
        - host
        - sub_domain
        - port
        - path
        - params
        - query
        - fragment
        - product_id
        """
        return ParseUrl(self.href)

    @property
    def response(self):
        retry_strategy = Retry(
            total=3,
            backoff_factor=0.2,
            status_forcelist=[502, 503, 504],
            method_whitelist=["GET"]
        )
        adapter = HTTPAdapter(max_retries=retry_strategy)
        session = requests.Session()
        for cookie in self.driver.get_cookies():
            session.cookies.set(cookie['name'], cookie['value'])
        session.mount("https://", adapter)
        session.mount("http://", adapter)
        return session.get(self.href, verify=False, timeout=10)

    @property
    def get_status_code(self):
        """Gets status requests code from link"""
        return self.response.status_code



from selenium.webdriver.common.by import By

from promium import Page, Block, Element
from tests.urls import collect_url


class FilterHintBlock(Block):

    close_hint_link = Element(
        By.CSS_SELECTOR,
        '[data-qaid="close_hint"]'
    )


class ProductTilBlock(Block):

    PRODUCT_ID_SELECTOR_ATTR = 'data-product-qaid'
    PRODUCT_POSITION_SELECTOR_ATTR = 'data-position-qaid'

    product_name = Element(
        By.CSS_SELECTOR,
        '[data-qaid="product_name"]'
    )
    presence = Element(
        By.CSS_SELECTOR,
        '[data-qaid="product_presence"]'
    )
    price = Element(
        By.CSS_SELECTOR,
        '[data-qaid="product_price"]'
    )
    old_price = Element(
        By.CSS_SELECTOR,
        '[data-qaid="old_price"]'
    )
    prosale_icon = Element(
        By.CSS_SELECTOR,
        '[data-qaid="prosale_icon"]'
    )
    add_favorite_icon = Element(
        By.CSS_SELECTOR,
        '[data-qaid="favorite"] svg'
    )
    flash_message = Element(
        By.CSS_SELECTOR,
        '[data-qaid="flash-message-text"]'
    )
    show_all_phones_link = Element(
        By.CSS_SELECTOR,
        '[data-qaid="show-all-phones-link"]'
    )


class ProductsBlock(Block):

    product_blocks = ProductTilBlock.as_list(
        By.CSS_SELECTOR,
        '[data-qaid="product_block"]'
    )

    def get_target_product_block(self, product_id):
        """
        Finds product block on products block by product id and return it.
        :param int product_id: "id" of the product which block you want to get.
        :return: product block
        """
        product_block = self.product_blocks.find(
            by=By.CSS_SELECTOR,
            locator='[{attribute}="{product_id}"]'.format(
                attribute=self.product_blocks.PRODUCT_ID_SELECTOR_ATTR,
                product_id=product_id
            )
        )
        return product_block


class CatalogGalleryPage(Page):

    url = collect_url('{query}')

    products_block = ProductsBlock(
        By.CSS_SELECTOR,
        '[data-qaid="product_gallery"]'
    )
    filter_hint_block = FilterHintBlock(
        By.CSS_SELECTOR,
        '[data-qaid="filter_hint"]'
    )
    show_more_button = Element(
        By.CSS_SELECTOR,
        '[data-qaid="show_more"]'
    )



from selenium.webdriver.common.by import By

from promium import Page, Element, InputField
from tests.urls import collect_url


class Catalog404Page(Page):

    url = collect_url('xxx')

    search_input = InputField(
        By.CSS_SELECTOR,
        '[data-qaid="search_form"] input'
    )
    search_button = Element(
        By.CSS_SELECTOR,
        '[data-qaid="search_form"] button[type="submit"]'
    )
    main_404_title = Element(
        By.CSS_SELECTOR,
        '[data-qaid="page_not_found_title"]'
    )
    suggest_404_title = Element(
        By.CSS_SELECTOR,
        '[data-qaid="page_not_found_descr"]'
    )




from promium.elements.input_field import InputField


class UploadButton(InputField):

    def upload_file(self, file_path):
        """Uploads file by file path"""
        self.send_keys(file_path)




import pytest
from selenium.webdriver.edge.options import Options as EdgeOptions

from promium.support.settings import SettingsEDGE


@pytest.mark.unit
def test_driver_capabilities():
    capabilities = SettingsEDGE.driver_capabilities
    assert capabilities
    assert isinstance(capabilities, dict)
    assert capabilities.get('browserName') == 'MicrosoftEdge'


@pytest.mark.unit
def test_binary_path():
    assert SettingsEDGE.binary_path
    assert 'edgedriver' in SettingsEDGE.binary_path


@pytest.mark.unit
def test_default_options():
    assert len(SettingsEDGE.default_options) >= 0
    for opt in SettingsEDGE.default_options:
        assert isinstance(opt, str)


@pytest.mark.unit
def test_driver_options():
    assert isinstance(SettingsEDGE.driver_options(), EdgeOptions)


@pytest.mark.unit
def test_get_options():
    assert isinstance(SettingsEDGE.get_options(), EdgeOptions)



import os
import datetime
import logging
import requests
from collections import namedtuple

import pytest
import allure

from requests.adapters import HTTPAdapter
from requests.packages.urllib3.util.retry import Retry

from selenium.common.exceptions import WebDriverException
from promium.test_case import create_driver
from promium.common import upload_screenshot
from promium.device_config import CHROME_DESKTOP_1920_1080, DesktopDevice
from promium.logger import Logger
from promium.exceptions import BrowserConsoleException, PromiumException


log = logging.getLogger(__name__)

FailInfoSelenium = namedtuple('FailInfoSelenium', [
    'test_type', 'url', 'screenshot', 'test_case', 'run_command', 'txid'
])
FailInfoRequest = namedtuple('FailInfoRequest', [
    'test_type', 'url', 'test_case', 'status_code', 'run_command', 'txid'
])


def pytest_sessionstart(session):
    if hasattr(session.config, 'cache'):
        cache = session.config.cache
        cache_path = "cache/{}".format(session.config.lastfailed_file)
        print("Lastfailed:", len(cache.get(cache_path, set())))
    session.run_duration = datetime.datetime.now()
    print("\nPytest session start %s\n" % session.run_duration)


@pytest.mark.trylast
def pytest_sessionfinish(session):
    finish = datetime.datetime.now()
    duration = datetime.timedelta(
        seconds=(finish - session.run_duration).seconds
    )
    print("\n\nPytest session finish %s (duration=%s)\n" % (finish, duration))


@pytest.hookimpl
def pytest_addoption(parser):
    parser.addoption(
        "--capturelog",
        dest="logger",
        default=None,
        action="store_true",
        help="Show log messages for each failed test"
    )
    parser.addoption(
        "--fail-debug-info",
        action="store_true",
        help="Show screenshot and test case urls for each failed test"
    )
    parser.addoption(
        "--duration-time",
        action="store_true",
        help="Show the very slow test"
    )
    parser.addoption(
        "--allure-report",
        action="store_true",
        help="Generate allure report"
    )
    parser.addoption(
        "--highlight",
        action="store_true",
        help="Highlighting elements"
    )
    parser.addoption(
        "--headless",
        action="store_true",
        help="Enable headless mode for Chrome browser only"
    )
    parser.addoption(
        "--chrome",
        action="store_true",
        help="Use chrome browser"
    )
    parser.addoption(
        "--firefox",
        action="store_true",
        help="Use firefox browser"
    )
    parser.addoption(
        "--opera",
        action="store_true",
        help="Use opera browser"
    )
    parser.addoption(
        "--repeat",
        action="store",
        default=1,
        type="int",
        metavar='repeat',
        help='Number of times to repeat each test. Mostly for debug purposes'
    )
    parser.addoption(
        "--check-console",
        action="store_true",
        help="Check browser console js and other errors"
    )
    parser.addoption(
        "--check-render-errors",
        action="store_true",
        help="Check browser console render errors"
    )
    parser.addoption(
        "-U",
        "--disable-test-case-url",
        action="store_false",
        help="If the flag is on, it turns off the mandatory test case"
    )
    parser.addoption(
        "-R",
        "--replace-headless",
        action="store",
        help=(
            """ When selenium tests running in Chrome Headless mode,
            the Headless browser can recognized as a bot,
            so we change the agent's username to a real one.
            By default user agent contains something like
            "HeadlessChrome/86.0.4240.75", but we need "Chrome/86.0.4240.75"
            You need to specify the Chrome version. """
        )
    )


@pytest.fixture()
def request_init(request):

    def _get_fail_debug(request):
        """Failed test report generator"""
        if hasattr(request.config, "get_failed_test_command"):
            run_command = request.config.get_failed_test_command(request)
        else:
            run_command = "Please, add run command in your fixture."

        session = request.cls.session
        if session:
            if not hasattr(session, 'status_code'):
                session.status_code = None

        return FailInfoRequest(
            test_type='request',
            url=(session.url, [None])[0],
            test_case=(request.cls.test_case_url, [None])[0],
            status_code=(session.status_code, [None])[0],
            run_command=(run_command, [None])[0],
            txid=(request.cls.xrequestid, [None])[0],
        )

    session = requests.session()
    retry_strategy = Retry(
        total=3,
        status_forcelist=[502, 503, 504],
        method_whitelist=["POST", "GET"]
    )
    adapter = HTTPAdapter(max_retries=retry_strategy)
    session.mount("https://", adapter)
    session.mount("http://", adapter)
    session.url = (
        'Use self.get_response(url) for request tests or '
        'util methods for api tests!'
    )
    if request.cls.proxies:
        session.proxies.update(request.cls.proxies)
    request.cls.session = session
    request.config.get_fail_debug = _get_fail_debug
    request.config.assertion_errors = request.cls.assertion_errors = []
    yield session
    if session:
        session.close()
    request.cls.xrequestid = None


@pytest.fixture()
def driver_init(request):
    def _get_screenshot_png():
        return driver.get_screenshot_as_png()

    def _get_fail_debug(request):
        """Failed test report generator"""
        driver = request.config.driver

        alerts = 0
        try:
            while driver.switch_to.alert:
                alert = driver.switch_to.alert
                print('Unexpected ALERT: %s\n' % alert.text)
                alerts += 1
                alert.dismiss()
        except Exception:
            if alerts != 0:
                print('')
            pass
        url = driver.current_url
        screenshot = upload_screenshot(driver)

        if hasattr(request.config, "get_failed_test_command"):
            run_command = request.config.get_failed_test_command(request)
        else:
            run_command = "Please, add run command in your fixture."

        return FailInfoSelenium(
            test_type='selenium',
            url=(url, [None])[0],
            screenshot=(screenshot, [None])[0],
            test_case=(request.cls.test_case_url, [None])[0],
            run_command=(run_command, [None])[0],
            txid=(request.cls.xrequestid, [None])[0],
        )

    def _check_render_errors(request):
        if hasattr(request.config.driver, "render_errors"):
            if request.config.driver.render_errors:
                return request.config.driver.render_errors
        return []

    def _is_skipped_error(request, error):
        for msg in request.cls.excluded_browser_console_errors:
            if msg["msg"] in error and msg["comment"]:
                return True
        return False

    def _check_console_errors(request):
        driver = request.config
        if hasattr(driver, "console_errors"):
            if driver.console_errors:
                browser_console_errors = driver.console_errors
                if hasattr(request.cls, "excluded_browser_console_errors"):
                    if request.cls.excluded_browser_console_errors:
                        try:
                            return list(filter(
                                lambda err: not _is_skipped_error(
                                    request, err),
                                browser_console_errors
                            ))
                        except Exception as e:
                            raise PromiumException(
                                "Please check your excluded errors list. "
                                "Original exception is: %s" % e
                            )
                return browser_console_errors
        return []

    mark = request.node.get_closest_marker('device')
    if mark:
        device = mark.args[0]
    else:
        device = CHROME_DESKTOP_1920_1080  # default data

    replacement_version = request.config.getoption('--replace-headless')
    if hasattr(device, "user_agent") and replacement_version:
        if not device.user_agent and "chrome" in os.environ.get("SE_DRIVER"):
            user_agent = (
                f"Mozilla/5.0 (X11; Linux x86_64) AppleWebKit/537.36 "
                f"(KHTML, like Gecko) "
                f"Chrome/{replacement_version} Safari/537.36"
            )
            device = DesktopDevice(1920, 1080, user_agent)

    request.cls.device = device
    if hasattr(request.config, "proxy_server"):
        proxy_server = request.config.proxy_server
    else:
        proxy_server = None
    driver = create_driver(device, proxy_server=proxy_server)
    driver.xrequestid = request.cls.xrequestid
    request.cls.driver = request.config.driver = driver
    request.config.get_fail_debug = _get_fail_debug
    request.config.get_screenshot_png = _get_screenshot_png
    request.config.check_console_errors = _check_console_errors
    request.config.assertion_errors = request.cls.assertion_errors = []
    request.config.render_errors = driver.render_errors = []
    request.config.check_render_errors = _check_render_errors
    request.config.console_errors = driver.console_errors = []
    yield driver
    driver.xrequestid = request.cls.xrequestid = None
    request.config.render_errors = driver.render_errors = []
    request.config.console_errors = driver.console_errors = []
    try:
        driver.quit()
    except WebDriverException as e:
        log.error(
            "[PROMIUM] Original webdriver exception: %s" % e
        )


def pytest_generate_tests(metafunc):
    if metafunc.config.option.repeat > 1:
        metafunc.fixturenames.append('repeat')
        metafunc.parametrize('repeat', range(metafunc.config.option.repeat))


@pytest.fixture(autouse=True)
def logger(request):
    return logging.getLogger()


@pytest.mark.trylast
def pytest_runtest_call(item):
    if hasattr(item.config, "assertion_errors"):
        if item.config.assertion_errors:
            separator = '-' * 46
            assertion_errors_list = f"{separator}\n".join(
                item.config.assertion_errors
            )
            msg = f"\n{separator}\n{assertion_errors_list}{separator}\n"
            raise AssertionError(msg)

    if item.config.getoption("--check-console"):
        if hasattr(item.config, "check_console_errors"):
            browser_console_errors = item.config.check_console_errors(item)
            if browser_console_errors:
                msg = (
                    "Browser console js errors found:"
                    "\n{console_errors}\n".format(
                        console_errors="\n".join(
                            browser_console_errors
                        )
                    )
                )
                raise BrowserConsoleException(msg)

    if item.config.getoption("--check-render-errors"):
        if hasattr(item.config, "check_render_errors"):
            render_errors = item.config.check_render_errors(item)
            if render_errors:
                msg = (
                    "Component render errors found:"
                    "\n{render_errors}\n".format(
                        render_errors="\n".join(
                            render_errors
                        )
                    )
                )
                raise BrowserConsoleException(msg)


@pytest.mark.tryfirst
@pytest.mark.hookwrapper
def pytest_runtest_makereport(item, call):
    """pytest failed test report generator"""
    html_plugin = item.config.pluginmanager.getplugin('html')
    outcome = yield
    report = outcome.get_result()
    when = report.when
    extra = getattr(report, 'extra', [])
    item.config.path_to_test = f'{item.location[0]} -k {item.name}'

    if when == "call" and not report.passed:
        if item.config.getoption("--fail-debug-info"):
            fail_info = "Fail info not found."
            try:
                if hasattr(item.config, "get_fail_debug"):
                    fail_debug = item.config.get_fail_debug(item)
                    if fail_debug.test_type == 'selenium':
                        fail_info = (
                            (
                                "\nURL: {url}"
                                "\nTEST CASE: {test_case}"
                                "\nSCREENSHOT: {screenshot}"
                                "\nRUN_COMMAND: {run_command}"
                                "\nTXID: {txid}"
                            ).format(
                                url=fail_debug.url,
                                screenshot=fail_debug.screenshot,
                                test_case=fail_debug.test_case,
                                run_command=fail_debug.run_command,
                                txid=fail_debug.txid
                            )
                        )
                        extra.append(html_plugin.extras.url(
                            fail_debug.screenshot, name="SCREENSHOT"
                        ))
                        extra.append(
                            html_plugin.extras.html(
                                f"<img src=\"{fail_debug.screenshot}\" "
                                f"height=\"230\" width=\"auto\" align=\""
                                f"right\" bottom=\"0px\">Screenshot</img>"
                            )
                        )
                        if item.config.getoption("--allure-report"):
                            allure.attach(
                                fail_debug.url, name='URL',
                                attachment_type='text/uri-list'
                            )
                            allure.attach(
                                item.config.get_screenshot_png(),
                                name='SCREENSHOT: OPEN ME, OPEN ME!',
                                attachment_type='image/png'
                            )

                    elif fail_debug.test_type == 'request':
                        fail_info = (
                            "\nURL: {url}"
                            "\nTEST_CASE: {test_case}"
                            "\nSTATUS_CODE: {status_code}"
                            "\nRUN_COMMAND: {run_command}"
                            "\nTXID: {txid}"
                        ).format(
                            url=fail_debug.url,
                            test_case=fail_debug.test_case,
                            status_code=fail_debug.status_code,
                            run_command=fail_debug.run_command,
                            txid=fail_debug.txid
                        )
                    extra.append(html_plugin.extras.url(
                        fail_debug.url, name="URL"
                    ))
                    extra.append(html_plugin.extras.url(
                        fail_debug.test_case, name="TEST_CASE"
                    ))
                    report.extra = extra
                    if item.config.getoption("--allure-report"):
                        if fail_debug.test_case:
                            allure.attach(
                                fail_debug.test_case,
                                name='TEST CASE',
                                attachment_type='text/uri-list'
                            )
                        allure.attach(
                            fail_debug.run_command,
                            name='RUN COMMAND'
                        )

            except WebDriverException as e:
                fail_info = (
                    "\nWebdriver instance must have crushed: {msg}".format(
                        msg=e.msg
                    )
                )
            finally:
                longrepr = getattr(report, 'longrepr', None)
                if hasattr(longrepr, 'addsection'):
                    longrepr.sections.insert(
                        0, ("Captured stdout %s" % when, fail_info, "-")
                    )

        if item.config.getoption("--duration-time"):
            duration = call.stop - call.start
            if duration > 120:
                report.sections.append((
                    "Captured stdout %s" % when,
                    ("\n\n!!!!! The very slow test. "
                     "Duration time is %s !!!!!\n\n")
                    % datetime.datetime.fromtimestamp(duration).strftime(
                        "%M min %S sec"
                    )
                ))


def pytest_configure(config):
    os.environ['PYTHONWARNINGS'] = (
        'ignore:An HTTPS request has been made, '
        'ignore:A true SSLContext object is not available, '
        'ignore:Unverified HTTPS request is being made'
    )
    os.environ['TEST_CASE'] = "True"

    if config.getoption('--headless'):
        os.environ['HEADLESS'] = 'Enabled'
    if config.getoption("--highlight"):
        os.environ['HIGHLIGHT'] = 'Enabled'
    if config.getoption("--capturelog"):
        config.pluginmanager.register(Logger(), "logger")
    if config.getoption("--check-console"):
        os.environ['CHECK_CONSOLE'] = 'Enabled'
    if config.getoption("--check-render-errors"):
        os.environ['CHECK_RENDER_ERRORS'] = 'Enabled'
    if not getattr(config, "lastfailed_file", None):
        config.lastfailed_file = "lastfailed"
    if config.getoption("--disable-test-case-url"):
        os.environ['TEST_CASE'] = "False"



import allure
import logging
import os
import py
import pytest
import re
from functools import wraps

from selenium.common.exceptions import WebDriverException


log = logging.getLogger(__name__)

MAX_SYMBOLS = 510
TABS_FORMAT = " " * 20


def repr_console_errors(console_errors, tabs=TABS_FORMAT):
    return "\n{tabs_format}".format(tabs_format=tabs).join(
        ">>> [CONSOLE ERROR] %s" % error for error in set(console_errors)
    )


def repr_args(*args, **kwargs):
    return "{args}{mark}{kwargs}".format(
        args=", ".join(list(map(lambda x: str(x), args))) if args else "",
        kwargs=", ".join(
            "%s=%s" % (k, v) for k, v in kwargs.items()
        ) if kwargs else "",
        mark=", " if args and kwargs else ""
    )


def find_console_browser_errors(driver):
    try:
        return set(map(
            lambda x: x["message"],
            list(filter(
                lambda x: x["level"] == "SEVERE", driver.get_log("browser")
            ))
        ))
    except WebDriverException as e:
        log.warning(f"Exception when reading Browser Console log: {str(e)}")
        return None


def find_component_render_errors(console_errors):
    return list(filter(lambda x: 'render error' in x, console_errors))


def logger_for_element_methods(fn):

    @wraps(fn)
    def wrapper(self, *args, **kwargs):
        res = fn(self, *args, **kwargs)
        console_errors = find_console_browser_errors(self.driver)
        is_check = os.environ.get('CHECK_CONSOLE')
        if console_errors:
            log.warning(
                "Browser console js error found:\n"
                "{tabs_format}{console_errors}\n"
                "{tabs_format}Url: {url}\n"
                "{tabs_format}Action: {class_name}"
                "({by}={locator})"
                ".{method}({args})".format(
                    tabs_format=TABS_FORMAT,
                    class_name=self.__class__.__name__,
                    by=self.by,
                    locator=self.locator,
                    method=fn.__name__,
                    args=repr_args(*args, **kwargs),
                    url=self.driver.current_url,
                    console_errors=repr_console_errors(
                        console_errors
                    )
                )
            )

            error_tmpl = (
                "Url: {url}\n"
                "Action: {class_name}({by}={locator})"
                ".{method}({args})\n"
                "{end_symbol}".format(
                    class_name=self.__class__.__name__,
                    by=self.by,
                    locator=self.locator,
                    method=fn.__name__,
                    args=repr_args(*args, **kwargs),
                    url=self.driver.current_url,
                    end_symbol="-" * 75
                )
            )
            is_check_render = os.environ.get('CHECK_RENDER_ERRORS')
            if is_check_render and hasattr(self.driver, "render_errors"):
                render_errors = find_component_render_errors(console_errors)
                for err in set(render_errors):
                    self.driver.render_errors.append(
                        f">>> [RENDER ERROR] {err}\n {error_tmpl}"
                    )
            if is_check and hasattr(self.driver, "console_errors"):
                for err in set(console_errors):
                    self.driver.console_errors.append(
                        f">>> [CONSOLE ERROR] {err}\n {error_tmpl}"
                    )
        return res
    return wrapper


def add_logger_to_base_element_classes(cls):
    for name, method in cls.__dict__.items():
        log.info("%s, %s" % (name, method))
        if (not name.startswith('_') and
                hasattr(method, '__call__') and name != "lookup"):
            setattr(cls, name, logger_for_element_methods(method))
    return cls


def logger_for_loading_page(fn):

    @wraps(fn)
    def wrapper(self, *args, **kwargs):
        res = fn(self, *args, **kwargs)
        if os.environ.get('CHECK_CONSOLE'):
            console_errors = find_console_browser_errors(self.driver)
            if console_errors:
                log.warning(
                    "Browser console js error found:\n"
                    "{tabs_format}{console_errors}\n"
                    "{tabs_format}Url: {url}\n"
                    "{tabs_format}Action: wait for page loaded ...".format(
                        tabs_format=TABS_FORMAT,
                        url=self.driver.current_url,
                        console_errors=repr_console_errors(console_errors)
                    )
                )
                if hasattr(self.driver, "console_errors"):
                    for err in set(console_errors):
                        self.driver.console_errors.append(
                            ">>> [CONSOLE ERROR] {err}\n"
                            "Url: {url}\n"
                            "Action: wait for page loaded ...\n"
                            "{end_symbol}".format(
                                url=self.driver.current_url,
                                err=err,
                                end_symbol="-" * 75
                            )
                        )
        return res
    return wrapper


IGNORE_LOG_MESSAGES = [
    "InsecureRequestWarning: Unverified HTTPS request is being made. "
    "Adding certificate verification is strongly advised.",
    "Starting new HTTP connection",
    "Starting new HTTPS connection"
]


class LoggerFilter(logging.Filter):

    def filter(self, record):
        if hasattr(record, "msg"):
            is_ignore = any(set(filter(
                lambda x: x in record.msg, IGNORE_LOG_MESSAGES
            )))
            if is_ignore:
                return False
        return record.levelno > 10


class Logger(object):

    def pytest_runtest_setup(self, item):
        item.capturelog_handler = LoggerHandler()
        item.capturelog_handler.setFormatter(logging.Formatter(
            "%(asctime)-12s%(levelname)-8s%(message)s\n", "%H:%M:%S"
        ))
        root_logger = logging.getLogger()
        item.capturelog_handler.addFilter(LoggerFilter())
        root_logger.addHandler(item.capturelog_handler)
        root_logger.setLevel(logging.NOTSET)

    @pytest.mark.hookwrapper
    def pytest_runtest_makereport(self, item, call):
        outcome = yield
        report = outcome.get_result()
        if hasattr(item, "capturelog_handler"):
            if call.when == 'teardown':
                root_logger = logging.getLogger()
                root_logger.removeHandler(item.capturelog_handler)
            if not report.passed:
                longrepr = getattr(report, 'longrepr', None)
                if hasattr(longrepr, 'addsection'):
                    captured_log = item.capturelog_handler.stream.getvalue()
                    if captured_log:
                        longrepr.sections.insert(
                            len(longrepr.sections),
                            ('Captured log', "\n%s" % captured_log, "-")
                        )
                        if item.config.getoption("--allure-report"):
                            allure.attach(
                                captured_log, name='CAPTURED LOG',
                                attachment_type='text/plain'
                            )
            if call.when == 'teardown':
                item.capturelog_handler.close()
                del item.capturelog_handler


class LoggerHandler(logging.StreamHandler):

    def __init__(self):
        logging.StreamHandler.__init__(self)
        self.stream = py.io.TextIO()
        self.records = []

    def close(self):
        logging.StreamHandler.close(self)
        self.stream.close()


def request_logging(request, *args, **kwargs):
    content = str(request.content)
    warning_message = (
        "[Python request] STATUS CODE: {status_code}, LINK: {link}\n"
        "{tabs}METHOD: {method}\n"
        "{tabs}BODY: {body}\n"
        "{tabs}RESPONSE CONTENT: "
        "{content}{ellipsis} ({length} symbols)\n".format(
            status_code=request.status_code,
            method=request.request.method,
            link=request.url,
            body=request.request.body,
            content=re.sub(r'\s+', ' ', content)[:MAX_SYMBOLS],
            ellipsis=" ..." if len(content) > MAX_SYMBOLS else "",
            length=len(request.content),
            tabs=" " * 12
        )
    )
    if 400 <= request.status_code < 500:
        log.warning(warning_message)
    elif request.status_code >= 500:
        log.error(warning_message)
    else:
        log.info(
            "[Python request] STATUS CODE: {status_code}, LINK: {link}".format(
                status_code=request.status_code, link=request.url
            )
        )



import errno
import os
import signal
import threading

import pytest
import logging
import traceback

from urllib.parse import urlsplit
from functools import wraps
from selenium.common.exceptions import WebDriverException

from promium.assertions import (
    WebDriverSoftAssertion,
    RequestSoftAssertion
)
from promium.exceptions import PromiumException
from promium.browsers import (
    FirefoxBrowser, ChromeBrowser, OperaBrowser, RemoteBrowser
)
from promium.logger import (
    request_logging,
    logger_for_loading_page
)


TEST_PROJECT = os.environ.get('TEST_PROJECT')
TEST_CASE = os.environ.get('TEST_CASE')

log = logging.getLogger(__name__)

DRIVERS = {
    'firefox': 'Firefox',
    'chrome': 'Chrome',
    'safari': 'Safari',
    'opera': 'Opera',
    'ie': 'Ie',
}

MAX_LOAD_TIME = 10


def timeout(seconds=5, error_message=os.strerror(errno.ETIME)):
    def decorator(func):
        def _handle_timeout(signum, frame):
            raise TimeoutError(error_message)

        @wraps(func)
        def wrapper(*args, **kwargs):
            if threading.current_thread() is threading.main_thread():
                signal.signal(signal.SIGALRM, _handle_timeout)
                signal.alarm(seconds)
            try:
                result = func(*args, **kwargs)
            finally:
                signal.alarm(0)
            return result
        return wrapper
    return decorator


@timeout(300)
def create_driver(
        device, proxy_server=None, env_var='SE_DRIVER', default='chrome://'
):
    """
    Examples:

        - 'chrome://'
        - 'firefox://'
        - 'opera://'

    """
    driver_options = {
        "device": device,
        "proxy_server": proxy_server,
        "is_headless": (
            True if os.environ.get("HEADLESS") == "Enabled" else False
        ),
    }

    driver_dsn = os.environ.get(env_var) or default

    if not driver_dsn:
        raise RuntimeError(f'Selenium WebDriver is not set in the {env_var} '
                           f'environment variable')
    try:
        scheme, netloc, url, _, _ = urlsplit(driver_dsn)
    except ValueError:
        raise ValueError(f'Invalid url: {driver_dsn}')

    if scheme in DRIVERS:
        if scheme == "chrome":
            return ChromeBrowser(**driver_options)
        elif scheme == "firefox":
            return FirefoxBrowser(device=device)
        elif scheme == "opera":
            return OperaBrowser(device=device)
        else:
            raise ValueError(f'Unknown client specified: {scheme}')

    elif scheme.startswith('http+'):
        proto, _, client = scheme.partition('+')
        if not netloc:
            raise ValueError(f'Network address is not specified: {driver_dsn}')

        if client not in DRIVERS:
            raise ValueError(f'Unknown client specified: {client}')

        driver_options["client"] = client
        driver_options["hub"] = f'{proto}://{netloc}{url}'

        try:
            driver = RemoteBrowser(**driver_options)
        except WebDriverException:
            log.warning("Second attempt for remote driver connection.")
            driver = RemoteBrowser(**driver_options)
        except KeyError as e:
            log.warning(str(e))
            if "status" in str(e):
                raise PromiumException(
                    "Session timed out because the browser was idle too long."
                    "Or no free slots for create new session."
                )
            raise
        return driver

    raise ValueError(f'Unknown driver specified: {driver_dsn}')


class TDPException(Exception):

    def __init__(self, *args):
        self.message = (
            "exception caught during execution test data preparing.\n"
            "Look at the original traceback:\n\n%s\n"
            ) % ("".join(traceback.format_exception(*args)))

    def __str__(self):
        return self.message


class TDPHandler:
    """
    TDP - Test Data Preparation
    context manager for handling any exception
    during execution test data preparing.
    We need to raise a specific custom exceptions.
    :example:
    with self.tdp_handler():
        some code
    """

    def __init__(self):
        pass

    def __enter__(self):
        log.info("[TDP] Start test data preparing...")
        return self

    def __exit__(self, exc_type, exc_value, exc_tb):
        log.info("[TDP] Finish test data preparing")
        if exc_type:
            raise TDPException(exc_type, exc_value, exc_tb)
        return


class TestCase(object):
    xrequestid = None
    test_case_url = None
    assertion_errors = None
    proxy_server = None

    def tdp_handler(self):
        """
        Use this context manager for prepare of test data only,
        not for business logic!
        """
        return TDPHandler()

    def setup_method(self, method):
        check_test_case = os.environ.get("TEST_CASE")
        if check_test_case == "True" and not self.test_case_url:
            raise PromiumException("Test don't have a test case url.")


@pytest.mark.usefixtures("driver_init")
class WebDriverTestCase(TestCase, WebDriverSoftAssertion):
    driver = None

    @logger_for_loading_page
    def get_url(self, url, cleanup=True):
        self.driver.get(url)
        if cleanup:
            try:
                self.driver.execute_script(
                    'localStorage.clear()'
                )
            except WebDriverException:
                pass
        return url


@pytest.mark.usefixtures("request_init")
class RequestTestCase(TestCase, RequestSoftAssertion):
    session = None
    proxies = {}

    def get_response(
        self, url, method="GET", timeout=10, cookies=None, **kwargs
    ):
        self.session.url = url
        self.session.status_code = None
        request_cookies = {'xrequestid': self.xrequestid}
        if cookies:
            request_cookies.update(cookies)
        response = self.session.request(
            method=method,
            url=url,
            timeout=timeout,
            verify=False,
            cookies=request_cookies,
            hooks=dict(response=request_logging),
            **kwargs,
        )
        self.session.status_code = response.status_code
        return response




import random

from promium.base import Element
from promium.base import Sequence
from promium.exceptions import PromiumException


class RadioButton(Element):

    def select(self):
        """Selects radio button"""
        if not self.is_selected():
            self.click()
        else:
            pass

    def select_random_button(self):
        """Selects random radio button from list"""
        if isinstance(self, Sequence):
            random.random(self).select()
        else:
            raise PromiumException("Element is not a list.")

    def is_one_button_already_selected(self):
        """Checks is one button from list already selected"""
        if isinstance(self, Sequence):
            return len(
                [button for button in self if button.is_selected()]
            ) == 1
        else:
            raise PromiumException("Element is not a list.")

    @property
    def label(self):
        """Gets label"""
        return self.driver.execute_script(
            """
            element = arguments[0];
            return element.parentElement.innerText;
            """,
            self.lookup()
        )




import logging
import os
import random
import time
import requests
from collections.abc import Sequence

from functools import wraps

from selenium.common.exceptions import (
    StaleElementReferenceException,
    WebDriverException,
    NoSuchElementException,
    TimeoutException
)
from selenium.webdriver.remote.webdriver import WebDriver
from selenium.webdriver.remote.webelement import WebElement
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.common.action_chains import ActionChains
from selenium.webdriver.common.by import By
from selenium.webdriver.support.wait import WebDriverWait

from promium.common import (
    switch_to_window, scroll_to_bottom, scroll_to_top,
    scroll_into_end, set_local_storage, find_network_log, scroll_to_element,
)
from promium.logger import (
    add_logger_to_base_element_classes,
    logger_for_loading_page
)
from promium.waits import (
    wait_for_page_loaded, wait_until, wait_for_animation,
    wait_document_status, wait_websocket_connection_open, wait_for_alert,
    wait_for_alert_is_displayed, wait_until_new_window_is_opened,
    wait_part_appear_in_class, wait_part_disappear_in_class,
)
from promium.expected_conditions import is_present
from promium.exceptions import (
    ElementLocationException,
    LocatorException,
    PromiumException
)


log = logging.getLogger(__name__)


MAX_TRIES = 20
NORMAL_LOAD_TIME = 10


def with_retries(tries, second_to_wait):
    """
    If element not found wait and try again
    If element found returned
    :param int tries:
    :param float second_to_wait: wait seconds
    :return: callable_func
    """

    def callable_func(func):

        @wraps(func)
        def wrapper(*args, **kwargs):
            for t in range(tries):
                res = func(*args, **kwargs)
                if res:
                    return res
                time.sleep(second_to_wait)

            element = args[0]
            raise ElementLocationException(
                f'Element with {element.by}={element.locator} not found'
            )

        return wrapper
    return callable_func


def highlight_element(func):
    """
    Enabled for debug only
    :param func:
    :return: wrapper
    """
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        element = func(self, *args, **kwargs)
        highlight_status = os.environ.get('HIGHLIGHT')
        if highlight_status == 'Enabled':
            self.driver.execute_script("""
                element = arguments[0];
                original_style = element.getAttribute('style');
                element.setAttribute('style', original_style +
                ";border: 3px solid #D4412B;");
                setTimeout(function(){
                element.setAttribute('style', original_style);
                    }, 300);
                """, element)
            time.sleep(.5)
        return element
    return wrapper


def get_web_driver(driver):
    if isinstance(driver, WebDriver):
        return driver
    elif isinstance(driver, WebElement):
        return get_web_driver(driver._parent)
    error = 'Wrong driver instance. Supported WebDriver and WebElement only'
    raise PromiumException(error)


class Page(object):

    def __init__(self, driver):
        self._driver = driver
        self.previous_window_handle = self.driver.current_window_handle

    url = None

    @property
    def driver(self):
        return self._driver

    def lookup(self):
        return None

    @logger_for_loading_page
    def open(self, **kwargs):
        if not self.url:
            raise PromiumException("Can't open page without url")
        start = time.time()
        try:
            self.driver.get(self.url.format(**kwargs))
        except TimeoutException:
            log.error(f"[PROMIUM] Page load time: {time.time() - start}")
            raise
        return self.url.format(**kwargs)

    @logger_for_loading_page
    def refresh(self):
        self.driver.refresh()

    def wait_for_page_loaded(self):
        wait_for_page_loaded(self.driver)

    def wait_for_source_changed(self, action_func, *args, **kwargs):
        source = self.driver.page_source
        action_func(*args, **kwargs)
        wait_until(
            self.driver, lambda driver: driver.page_source != source,
            msg="Page source didn't change."
        )

    def get_status_code(self):
        """ get status_code from current page """
        session = requests.Session()
        for cookie in self.driver.get_cookies():
            session.cookies.set(cookie['name'], cookie['value'])
        r = session.get(
            url=self.driver.current_url,
            timeout=5, verify=False, allow_redirects=False
        )
        session.close()
        return r.status_code

    def wait_for_animation(self):
        wait_for_animation(self.driver)

    def wait_document_status(self, seconds=NORMAL_LOAD_TIME):
        wait_document_status(self.driver, seconds=seconds)

    def wait_websocket_connection_open(self, seconds=NORMAL_LOAD_TIME):
        wait_websocket_connection_open(self.driver, seconds=seconds)

    def wait_for_alert(self, seconds=NORMAL_LOAD_TIME):
        return wait_for_alert(self.driver, seconds=seconds)

    def is_alert_displayed(self, seconds=NORMAL_LOAD_TIME):
        return wait_for_alert_is_displayed(self.driver, seconds=seconds)

    def switch_to_tab(self):
        new_window = wait_until_new_window_is_opened(
            self.driver, self.previous_window_handle
        )
        switch_to_window(self.driver, new_window)

    def switch_previous_tab(self):
        switch_to_window(self.driver, self.previous_window_handle)

    def scroll_to_bottom(self):
        scroll_to_bottom(self.driver)

    def scroll_to_top(self):
        scroll_to_top(self.driver)

    def scroll_into_end(self):
        scroll_into_end(self.driver)

    def find_network_log(self, find_mask, find_status=200, timeout=5):
        return find_network_log(self.driver, find_mask, find_status, timeout)

    def set_local_storage(self, key, value):
        set_local_storage(self.driver, key, value)


class Bindable(object):

    def bind(self, parent, driver, parent_cls=None):
        c = self.__class__.__new__(self.__class__)
        c.__dict__ = self.__dict__.copy()
        c.parent = parent
        c.parent_cls = parent_cls
        c.driver = driver
        return c

    def __get__(self, instance, owner):
        return self.bind(instance.lookup(), instance.driver, instance)


class Base(Bindable):

    def __init__(self, by, locator):
        self.by = by
        self.locator = locator

    index = 0
    parent = None
    driver = None
    parent_cls = None
    _cached_elements = None

    def _refresh_parent(self):
        if self.parent_cls:
            self.driver = self.parent_cls.driver
            self.parent = self.parent_cls.lookup()

    def _get_current_driver(self):
        """
        Returned WebDriver or WebElement
        :return: obj
        """
        obj = self.parent if self.parent is not None else self.driver
        return obj

    def _validate_locator(self):
        """
        Validate XPATH locators:
        locator mast have start "."
        for example ".//*[@id='cabinet']"
        :return tuple[str, str]
        """
        if isinstance(self._get_current_driver(), WebElement):
            if self.by == By.XPATH and not self.locator.startswith("."):
                raise LocatorException(
                    f'Xpath locator must start with "." (dot). '
                    f'Please check this locator "{self.locator}"'
                )
        return self.by, self.locator

    def _find_elements(self):
        driver = self._get_current_driver()
        locator = self._validate_locator()
        try:
            self._cached_elements = driver.find_elements(*locator)
        except StaleElementReferenceException:
            self._refresh_parent()
            self._cached_elements = driver.find_elements(*locator)
        return self._cached_elements

    @with_retries(MAX_TRIES, .5)
    def _finder(self, force=False):
        """
        Used cache, find element only first call
        :return: self._cached_elements
        """
        if force or not self._cached_elements:
            self._find_elements()
        return self._cached_elements


class Collection(Base, Sequence):

    def __iter__(self):
        for i in range(len(self)):
            yield self[i]

    def get_item(self, index):
        item_name = self.base.__name__ + 'Item'
        params = {
            'parent': self.parent,
            'driver': self.driver,
            'parent_cls': self.base.parent_cls,
            '_cached_elements': self._cached_elements,
            'index': index
        }
        return type(item_name, (self.base, ), params)(self.by, self.locator)

    def __getitem__(self, val):
        if isinstance(val, slice):
            start = val.start or 0
            stop = val.stop if val.stop < len(self) else len(self)
            step = val.step or 1
            return [self.get_item(i) for i in range(start, stop, step)]

        return self.get_item(val)

    def __len__(self):
        return len(self._find_elements())

    def is_empty(self):
        return bool(len(self) < 1)

    def is_not_empty(self):
        return not self.is_empty()

    def random_choice(self):
        elements = self.wait_not_empty()
        index = random.randint(0, len(elements) - 1)
        return self.get_item(index)

    def wait_empty(self, timeout=10):
        msg = 'Collection with {}={} is not empty after {}'.format(
            self.by, self.locator, timeout
        )
        wait_until(self.driver, lambda e: self.is_empty(), timeout, msg)
        return self

    def wait_not_empty(self, timeout=10):
        msg = 'Collection with {}={} is empty after {}'.format(
            self.by, self.locator, timeout
        )
        wait_until(self.driver, lambda e: self.is_not_empty(), timeout, msg)
        return self

    def filter_by(self, child_locator):
        """
        Filtered blocks contained child CSS locator for example:
        Get product_blocks have buy_button
        :param str child_locator: example [data-qaid="buy_button"]
        """
        script = "return $('{parent_locator}:has({child_locator})');".format(
            parent_locator=self.locator,
            child_locator=child_locator
        )
        results = self.driver.execute_script(script)
        if not results:
            ElementLocationException(
                f'Not found elements with {child_locator}'
            )
        self._cached_elements = results
        return self

    def find_all(self, by, locator):
        """
        Changed locator for lookup
        :param str by: example By.CSS_SELECTOR
        :param str locator: example [data-qaid="some-locator"]
        :return: Collection
        """
        c = self.bind(self.parent, self.driver, self)
        c.by = by
        c.locator = locator
        return c

    def find(self, by, locator):
        """
        Changed locator for lookup
        :param str by: example By.CSS_SELECTOR
        :param str locator: example [data-qaid="some-locator"]
        :return: Element
        """
        return self.find_all(by, locator).get_item(0)

    def find_by_text(self, text, check_negative=False,  partial_text=True):
        found_elements_by_test = []
        if partial_text:
            script = """
                var el = arguments[0];
                var text = arguments[1];
                var element
                var index;
                if (el.innerText.includes(text)) element = true;
                return element
                """
        else:
            script = """
                var el = arguments[0];
                var text = arguments[1];
                var element
                var index;
                if (el.innerText == text) element = true;
                return element
            """
        for element in self:
            if self.driver.execute_script(script, element.lookup(), text):
                found_elements_by_test.append(element)
        if len(found_elements_by_test) == 0:
            if check_negative:
                return False
            else:
                raise PromiumException(
                    f'Element with text "{text}" has not been found'
                )
        else:
            return found_elements_by_test[0]

    @property
    def first_item(self):
        return self.wait_not_empty().get_item(0)

    @property
    def last_item(self):
        return self.wait_not_empty().get_item(-1)


@add_logger_to_base_element_classes
class ElementBase(Base):

    def _get_element(self):
        if self.index >= len(self._cached_elements):
            raise PromiumException(
                'Collection by={by} locator={locator} has been changed, '
                'get {index} of {maximum} elements'.format(
                    by=self.by,
                    locator=self.locator,
                    index=self.index,
                    maximum=len(self._cached_elements) - 1
                )
            )
        return self._cached_elements[self.index]

    @highlight_element
    def lookup(self):
        try:
            self._cached_elements = self._finder()
            element = self._get_element()
            element.is_displayed()
        except StaleElementReferenceException:
            self._refresh_parent()
            self._cached_elements = self._finder(force=True)
            element = self._get_element()
        return element

    @classmethod
    def as_list(cls, by, locator):
        params = {'parent': cls.parent, 'driver': cls.driver, 'base': cls}
        return type(cls.__name__ + 'List', (Collection, ), params)(by, locator)

    def wait_to_display(self, timeout=10, msg=''):
        """
        Waited to display element in the visible area of the page.
        :param int timeout:
        :param str msg:
        :return: Element
        """
        obj = self._get_current_driver()
        try:
            WebDriverWait(
                obj, timeout, ignored_exceptions=[
                    WebDriverException,
                    ElementLocationException
                ]
            ).until(lambda driver: self.is_displayed())
            return self
        except TimeoutException:
            if self.is_present():
                msg = (
                    f'Element with {self.by}={self.locator} was found '
                    f'but was not displayed after {timeout} seconds'
                    f'\n{msg}'
                )
            else:
                msg = (
                    f'Element with {self.by}={self.locator} not found '
                    f'after {timeout} seconds'
                    f'\n{msg}'
                )
            raise ElementLocationException(msg)

    def wait_to_present(self, timeout=10, msg=''):
        """
        Waited to present element in DOM.
        :param int timeout:
        :param str msg:
        :return: Element
        """
        wait_until(
            self.driver,
            is_present(self),
            timeout,
            f'Element with {self.by}={self.locator} not found '
            f'after {timeout} seconds'
            f'\n{msg}'
        )
        return self

    def wait_to_disappear(self, time=10, msg=''):
        def _find_no_visible_element(self):
            obj = self._get_current_driver()
            try:
                return not obj.find_element(
                    self.by, self.locator
                ).is_displayed()
            except (
                NoSuchElementException,
                StaleElementReferenceException,
                WebDriverException
            ):
                return True
        return WebDriverWait(self.driver, time).until(
            lambda driver: _find_no_visible_element(self),
            msg if msg else
            f'Element with {self.by}={self.locator} did not disappear '
            f'after {time} seconds'
        )

    def wait_in_viewport(self, time=5):
        jquery_script = """
            var elem = arguments[0],
            box = elem.getBoundingClientRect(),
            cx = box.left + box.width / 2,
            cy = box.top + box.height / 2,
            e = document.elementFromPoint(cx, cy);
            for (; e; e = e.parentElement) {
                if (e === elem)
                    return true;
            }
            return false;
            """

        wait_until(
            driver=self.driver,
            expression=lambda driver:
            driver.execute_script(jquery_script, self.lookup()),
            seconds=time,
            msg=f'Element with {self.by}={self.locator} did not appear '
                f'in viewport after {time} seconds'
            )
        return self

    def wait_part_appear_in_class(self, part_class, msg=None):
        wait_part_appear_in_class(
            driver=self.driver, obj=self.lookup(),
            part_class=part_class, msg=msg,
        )

    def wait_part_disappear_in_class(self, part_class, msg=None):
        wait_part_disappear_in_class(
            driver=self.driver, obj=self.lookup(),
            part_class=part_class, msg=msg,
        )

    def scroll(self, base_element=None):
        scroll_to_element(self.driver, self.lookup(), base_element)

    def get_attribute(self, name):
        return self.lookup().get_attribute(name)

    def check_class_part_presence(self, class_name):
        return bool(class_name in self.lookup().get_attribute('class'))

    @property
    def location(self):
        return self.lookup().location

    def is_present(self):
        """Checks if element is present on page at current time"""
        try:
            return bool(self._find_elements())
        except WebDriverException as e:
            log.error(f"[PROMIUM] Original webdriver exception: {e}")
            return False

    @property
    def text(self):
        return self.lookup().text

    def click(self):
        """ custom click, with blackjack and hookers """
        self.scroll_into_view()
        self.lookup().click()

    def js_click(self):
        self.driver.execute_script('arguments[0].click()', self.lookup())

    def is_displayed(self):
        return self.lookup().is_displayed()

    def hover_over(self):
        ActionChains(self.driver).move_to_element(self.lookup()).perform()

    def hover_over_with_offset(self, xoffset=1, yoffset=1):
        ActionChains(self.driver).move_to_element_with_offset(
            self.lookup(), xoffset, yoffset
        ).perform()

    def click_with_offset(self, xoffset=1, yoffset=1):
        ActionChains(self.driver).move_to_element_with_offset(
            self.lookup(), xoffset, yoffset
        ).click().perform()

    def move_to_element_and_click(self):
        ActionChains(self.driver).move_to_element(
            self.lookup()).click().perform()

    @property
    def element_title(self):
        return self.get_attribute("title")

    @property
    def element_id(self):
        return self.get_attribute("id")

    def value_of_css_property(self, property_name):
        """ Returns the value of a CSS property """
        return self.lookup().value_of_css_property(property_name)

    def scroll_into_view(
        self, behavior="instant", block="center", inline="center"
    ):
        """
        Method scrolls the element on which it's called into the visible
        area of the browser window.

        element.scrollIntoView();
        element.scrollIntoView(scrollIntoViewOptions); // Object parameter

        :Args:
        - behavior - Defines the transition animation. One of "auto" or
            "smooth" or "instant".
        - block - Defines vertical alignment. One of "start", "center",
            "end", or "nearest".
        - inline - Defines horizontal alignment. One of "start", "center",
            "end", or "nearest".
        """
        options = {
            "behavior": behavior,
            "block": block,
            "inline": inline
        }
        self.driver.execute_script(
            'arguments[0].scrollIntoView(arguments[1]);',
            self.lookup(), options
        )
        return self

    def drag_and_drop(self, target):
        ActionChains(self.driver).drag_and_drop(
            self.lookup(), target.lookup()
        ).perform()


class Block(ElementBase):
    pass


class Element(ElementBase):

    @property
    def tag_name(self):
        """Gets this element's tagName property."""
        return self.lookup().tag_name

    def submit(self):
        """Submits a form."""
        return self.lookup().submit()

    def is_selected(self):
        """
        Can be used to check if a checkbox or radio button is selected.
        """
        return self.lookup().is_selected()

    def is_enabled(self):
        """Whether the element is enabled."""
        return self.lookup().is_enabled()

    @property
    def location_once_scrolled_into_view(self):
        """CONSIDERED LIABLE TO CHANGE WITHOUT WARNING.
        Use this to discover where on the screen an element is
        so that we can click it. This method should cause
        the element to be scrolled into view.

        Returns the top lefthand corner location on the screen,
        or None if the element is not visible
        """
        return self.lookup().location_once_scrolled_into_view

    @property
    def size(self):
        """ Returns the size of the element """
        return self.lookup().size

    @property
    def id(self):
        """ Returns internal id used by selenium.

        This is mainly for internal use.  Simple use cases such as checking
        if 2 webelements refer to the same element, can be done using '=='::

            if element1 == element2:
                print("These 2 are equal")

        """
        return self.lookup().id

    def wait_to_staleness(self):
        return wait_until(
            driver=self.driver,
            expression=EC.staleness_of(self.lookup()),
            seconds=10,
        )

    def wait_frame_and_switch_to_it(self):
        return wait_until(
            driver=self.driver,
            expression=(
                EC.frame_to_be_available_and_switch_to_it(self.lookup())
            ),
            seconds=10
        )

    def wait_for_text_present(self, text):
        params = (self.by, self.locator)
        return wait_until(
            driver=self.driver,
            seconds=10,
            expression=EC.text_to_be_present_in_element(params, text),
            msg=f"Text: {text}, does not appear after 10 seconds"
        )

    def set_inner_html(self, value):
        self.driver.execute_script(
            'arguments[0].innerHTML = "%s"' % value, self.lookup()
        )

    def hover_over_and_click(self, on_element=None):
        ActionChains(self.driver).move_to_element(self.lookup()).click(
            on_element.lookup()
        ).perform()

    def drag_and_drop_by_offset(self, xoffset, yoffset):
        ActionChains(self.driver).drag_and_drop_by_offset(
            self.lookup(), xoffset, yoffset
        ).perform()

    def wait_for_changes_text_attribute(self, name_attr, text):
        return wait_until(
            driver=self.driver,
            expression=lambda driver: text != self.get_attribute(name_attr),
            seconds=10,
            msg='Text attribute has not changed'
        )

    def wait_for_text_change(self, text=None):
        element_text = text if type(text) is str else self.text
        return wait_until(
            driver=self.driver,
            expression=lambda driver: element_text != self.text,
            seconds=5,
            msg='Text "{}" not changed'.format(element_text)
        )

    def wait_to_enabled(self, time=10):
        try:
            wait = WebDriverWait(
                self.driver, time,
                ignored_exceptions=[WebDriverException])
            message = "Element with %s=%s not enabled after %s seconds" % (
                self.by, self.locator, time
            )
            return wait.until(
                lambda driver:
                False if self.get_attribute("disabled") == 'true'
                else self, message
            )
        except TimeoutException as e:
            raise ElementLocationException(e.msg)




import pytest

from promium.support.settings import SettingsOpera


@pytest.mark.unit
def test_driver_capabilities():
    assert SettingsOpera.driver_capabilities
    assert isinstance(SettingsOpera.driver_capabilities, dict)
    assert SettingsOpera.driver_capabilities.get('browserName') == 'opera'


@pytest.mark.unit
def test_binary_path():
    assert SettingsOpera.binary_path
    assert 'opera' in SettingsOpera.binary_path




import os
import datetime
import urllib
import requests
import json
import time
import logging

from selenium.webdriver.common.keys import Keys
from selenium.webdriver.common.action_chains import ActionChains

from promium.waits import (
    wait_for_animation, enable_jquery, wait_until_new_window_is_opened
)


log = logging.getLogger(__name__)

SCREENSHOT_SERVER_HOST = os.environ.get("SCREENSHOT_SERVER_HOST")
SCREENSHOT_SERVER_LOGIN = os.environ.get("SCREENSHOT_SERVER_LOGIN")
SCREENSHOT_SERVER_PASSWORD = os.environ.get("SCREENSHOT_SERVER_PASSWORD")


def scroll_to_bottom(driver):
    """Scrolls down page"""
    enable_jquery(driver)
    driver.execute_script('jQuery("img.img-ondemand").trigger("appear");')
    driver.execute_script(
        """
        var f = function(old_height) {
            var height = $$(document).height();
            if (height == old_height) return;
            $$('html, body').animate({scrollTop:height}, 'slow', null,
            setTimeout(function() {f(height)}, 1000)
            );
        }
        f();
        """
    )
    wait_for_animation(driver)


def scroll_to_top(driver):
    enable_jquery(driver)
    driver.execute_script(
        """jQuery('html, body').animate({scrollTop: 0 }, 'slow', null);"""
    )
    wait_for_animation(driver)


def scroll_to_bottom_in_block(driver, element_class):
    enable_jquery(driver)
    script = """
        var elem = '.'.concat(arguments[0]);
        $(elem).animate({scrollTop: $(elem).prop('scrollHeight')}, 1000);
    """
    driver.execute_script(script, element_class)
    wait_for_animation(driver)


def scroll_to_element(driver, element, base_element=None):
    """
    use base_element if you need for example scroll into popup,
    base_element must be a popup locator.
    """
    enable_jquery(driver)
    if base_element is None:
        base_element = 'html, body'
    script = """
        var elem = arguments[0];
        var base = arguments[1];
        var relativeOffset = (
            jQuery(elem).offset().top - jQuery(base).offset().top
        );
        jQuery(base).animate({
            scrollTop: relativeOffset
            }, 'slow', null
        );
             """
    driver.execute_script(script, element, base_element)
    wait_for_animation(driver)


def scroll_with_offset(driver, element, with_offset=0):
    enable_jquery(driver)
    script = """
        var elem = arguments[0];
        jQuery('html, body').animate({
            scrollTop: jQuery(elem).offset().top + arguments[1]
            }, 'fast', null
        );
        """
    driver.execute_script(script, element, with_offset)
    wait_for_animation(driver)


def scroll_into_end(driver):
    """Scroll block into end page"""
    enable_jquery(driver)
    driver.execute_script(
        'window.scrollTo(0, document.body.scrollHeight);'
    )
    wait_for_animation(driver)


def open_link_in_new_tab(driver, url=None):
    """
    Opening URL in new tab and switches to recently opened window
    If URL is None - an empty window will open
    """
    main_window = driver.current_window_handle
    driver.execute_script(
        f'''window.open("{
        "about:blank" if url is None else url
        }","_blank");'''
    )
    new_window = wait_until_new_window_is_opened(driver, main_window)
    switch_to_window(driver, new_window)


# TODO need implemented
def switch_to_window(driver, window_handle):
    """Switches to window"""
    driver.switch_to.window(window_handle)


def get_screenshot_path(name, with_date=True):
    now = datetime.datetime.now().strftime('%d_%H_%M_%S_%f')
    screenshot_name = f"{name}_{now}.png" if with_date else f"{name}.png"
    screenshots_folder = "/tmp"
    if not os.path.exists(screenshots_folder):
        os.makedirs(screenshots_folder)
    screenshot_path = os.path.join(screenshots_folder, screenshot_name)
    return screenshot_path, screenshot_name


def _upload_to_server(screenshot_path, screenshot_name):

    def read_in_chunks(img, block_size=1024, chunks=-1):
        """
        Lazy function (generator) to read a file piece by piece.
        Default chunk size: 1k.
        """
        while chunks:
            data = img.read(block_size)
            if not data:
                break
            yield data
            chunks -= 1

    screenshot_url = f"https://{SCREENSHOT_SERVER_HOST}/{screenshot_name}"
    r = requests.put(
        url=screenshot_url,
        auth=(SCREENSHOT_SERVER_LOGIN, SCREENSHOT_SERVER_PASSWORD),
        data=read_in_chunks(open(screenshot_path, 'rb')),
        verify=False
    )
    os.remove(screenshot_path)
    if not r.ok:
        return (
            f"Screenshot not uploaded to server. "
            f"Status code: {r.status_code}, reason: {r.reason}"
        )
    return screenshot_url


def decode_and_upload_screen(screen):
    try:
        import base64
        screenshot_path, screenshot_name = get_screenshot_path(
            name="timeout_screen"
        )
        with open(screenshot_path, "wb") as fh:
            fh.write(base64.decodebytes(screen.encode()))

        if SCREENSHOT_SERVER_HOST:
            return _upload_to_server(screenshot_path, screenshot_name)
        return f"file://{screenshot_path}"
    except Exception as e:
        return f"No decode screen due to exception {repr(e)}"


def upload_screenshot(driver, path_name="screenshot", path_with_date=True):
    try:
        screenshot_path, screenshot_name = get_screenshot_path(
            name=path_name,
            with_date=path_with_date
        )
        driver.save_screenshot(screenshot_path)

        if SCREENSHOT_SERVER_HOST:
            return _upload_to_server(screenshot_path, screenshot_name)
        return f"file://{screenshot_path}"

    except Exception as e:
        return f'No screenshot was captured due to exception {repr(e)}'


def download_and_check_file(download_url, file_path):
    try:
        urllib.urlretrieve(download_url, file_path)
        if os.path.isfile(file_path) and os.path.exists(file_path):
            return True
        else:
            return False
    finally:
        if os.path.exists(file_path):
            os.remove(file_path)


def control_plus_key(driver, key):
    """Imitations press CONTROL key + any key"""
    (
        ActionChains(driver)
        .key_down(Keys.CONTROL)
        .send_keys(key)
        .key_up(Keys.CONTROL)
        .perform()
    )


def set_local_storage(driver, key, value):
    """Sets value in browsers local storage"""
    driver.execute_script(
        "localStorage.setItem('%s', '%s')" % (key, value)
    )


def delete_cookie_item(driver, name):
    driver.delete_cookie(name)
    driver.refresh()


def delete_element_from_dom(driver, element):
    enable_jquery(driver)
    driver.execute_script(f"""
        var element = document.querySelector("{element}");
        if (element)
        element.parentNode.removeChild(element);
    """)


def find_network_log(driver, find_mask, find_status=200, timeout=5):
    """ find xrh response and requset data from browser network logs
    :param driver: selenium driver
    :param find_mask: find word in request url
    :param find_status: find response with this status
    :param timeout: tries
    :return: {response, data, body} or None
    """
    found_logs = []

    def get_log(request_mask):
        for performance_log in driver.get_log('performance'):
            # only response logs
            if (
                request_mask in performance_log['message'] and
                'responsereceived' in performance_log['message'].lower()
            ):
                json_log = json.loads(performance_log['message'])['message']
                if not json_log['params'].get('response'):
                    continue
                response = json_log['params']['response']
                found_logs.append({
                    'url': response.get('url'),
                    'status': (
                        f"{response['status']} - {response['statusText']}"
                    )
                })
                if response['status'] == find_status:
                    return json_log

    while timeout >= 0:
        xhr_log = get_log(find_mask)
        if xhr_log:
            requestId = xhr_log['params']['requestId']
            response = xhr_log['params']['response']
            break
        time.sleep(0.5)
        timeout -= 1

    try:
        request_data = driver.execute_cdp_cmd(
            'Network.getRequestPostData', {'requestId': requestId}
        )
        request_data = json.loads(request_data['postData'])
    except Exception:
        request_data = None

    try:
        response_body = driver.execute_cdp_cmd(
            'Network.getResponseBody', {'requestId': requestId}
        )
        response_body = json.loads(response_body['body'])
    except Exception:
        response_body = None

    log.info(
        f"[XHR Logs]: find by mask: '{find_mask}', result: {found_logs}, "
        f"request_data: {request_data}, response_body: {response_body},"
    )

    if xhr_log:
        return {
            'response': {'url': response['url'], 'status': response['status']},
            'data': request_data,
            'body': response_body,
        }



# -*- coding: utf-8 -*-

import pytest

from promium.test_case import WebDriverTestCase
from promium.waits import wait_url_contains
from tests.urls import collect_url
from tests.pages.catalog_404_page import Catalog404Page


MAIN_404_PAGE_TITLE = (
    'Оо! Страница не найдена'
)
SUGGEST_404_PAGE_TITLE = (
    'Похоже такой страницы больше нет, либо произошла какая-то ошибка, '
    'попробуйте обновить страницу либо поискать что-то еще'
)

SEARCH_TERM = 'search?search_term='


def get_expected_search_term_url(search_target):
    return collect_url('{search_term}{search_target}'.format(
        search_term=SEARCH_TERM,
        search_target=search_target
    ))


@pytest.mark.se
class Test404CatalogPage(WebDriverTestCase):
    test_case_url = 'some url with test case'

    def test_catalog_404_page_check_elements(self):
        catalog_404_page = Catalog404Page(self.driver)
        catalog_404_page.open()
        catalog_404_page.wait_for_page_loaded()
        self.soft_assert_equals(
            catalog_404_page.main_404_title.text,
            MAIN_404_PAGE_TITLE,
            'Не правильно указан основной title на странице 404.'
        )
        self.soft_assert_equals(
            catalog_404_page.suggest_404_title.text,
            SUGGEST_404_PAGE_TITLE,
            'Не правильно указан suggest title на странице 404.'
        )
        fake_search_query = 666
        catalog_404_page.search_input.send_keys(fake_search_query)
        catalog_404_page.search_button.click()
        wait_url_contains(
            self.driver,
            get_expected_search_term_url(fake_search_query)
        )



import logging

# TODO mast be yml config
from tests.config import PROJECT_HOST, PROJECT_PORT, WIKI_HOST


log = logging.getLogger(__name__)


def collect_url(path, sub_domain=''):
    """
    Collects url
    :param str path:
    :param str sub_domain:
    :return str url
    """
    return 'https://{sub_domain}{domain}{port}/{path}'.format(
        path=path,
        domain=PROJECT_HOST,
        port=PROJECT_PORT,
        sub_domain=sub_domain
    )


def case_url(id):
    """
    Returns check-list page by id
    :param int id: Page id
    :return str url
    """
    return 'https://{wiki_host}{id}'.format(
        wiki_host=WIKI_HOST,
        id=id
    )




from promium.exceptions import PromiumException


class BaseElementConditionWithRefresh(object):

    def __init__(self, element, with_refresh=False):
        self.element = element
        self.with_refresh = with_refresh

    def call_method(self):
        raise PromiumException('Not implemented')

    def __call__(self, driver):
        if self.with_refresh:
            driver.refresh()
        return self.call_method()


class ElementPresent(BaseElementConditionWithRefresh):

    def call_method(self):
        return bool(self.element.is_present())


is_present = ElementPresent


class ElementNotPresent(BaseElementConditionWithRefresh):

    def call_method(self):
        return bool(not self.element.is_present())


is_not_present = ElementNotPresent


class ElementIsDisplay(BaseElementConditionWithRefresh):

    def call_method(self):
        return bool(self.element.is_displayed())


is_display = ElementIsDisplay


class ElementNotDisplay(BaseElementConditionWithRefresh):

    def call_method(self):
        return bool(not self.element.is_displayed())


is_not_display = ElementNotDisplay


class ElementText(BaseElementConditionWithRefresh):

    def __init__(self, element, text, with_refresh=False):
        super(ElementText, self).__init__(element, with_refresh)
        self.text = text

    def call_method(self):
        return bool(self.element.text == self.text)


is_text = ElementText


class ElementTextContain(ElementText):

    def call_method(self):
        return bool(self.text in self.element.text)


is_text_contain = ElementTextContain


class ElementTextNotContain(ElementText):

    def bool_method(self):
        return bool(self.text not in self.element.text)


is_text_not_contain = ElementTextNotContain



from selenium.webdriver.common.by import By

from promium import Page, Block, Link, InputField

from tests.urls import collect_url


class HeaderRegistration(Block):
    pass


class TopSearchesBlock(Block):
    pass


class EmptyOrderPopup(Block):
    pass


class SearchBlock(Block):
    search_input = InputField(
        By.CSS_SELECTOR,
        '[data-qaid="search_input"]'
    )


class HeaderBlock(Block):

    search_block = SearchBlock(
        By.CSS_SELECTOR,
        '[data-qaid="search_input_block"]'
    )
    logo_link = Link(
        By.CSS_SELECTOR,
        '[data-qaid="header_logo_link"]'
    )
    favourite_link = Link(
        By.CSS_SELECTOR,
        '[data-qaid="favorite"]'
    )
    shopping_cart_link = Link(
        By.CSS_SELECTOR,
        '[data-qaid="shopping_cart"]'
    )
    authorization_link = Link(
        By.CSS_SELECTOR,
        '[data-qaid="auth_element"][href*="sign-in"]'
    )


class HeaderToolbarBlock(Block):

    add_classified_link = Link(
        By.CSS_SELECTOR,
        '[data-qaid="add_classified"]'
    )
    mobile_app_link = Link(
        By.CSS_SELECTOR,
        '[data-qaid="mobile_app_link"]'
    )
    authorization_link = Link(
        By.CSS_SELECTOR,
        '[data-qaid="auth_element"]'
    )
    registration_link = Link(
        By.CSS_SELECTOR,
        "[data-qaid='reg_element']"
    )
    header_reg_popup = HeaderRegistration(
        By.CSS_SELECTOR,
        '[data-qaid="header_registration_popup"]'
    )


class CatalogMainPage(Page):

    url = collect_url('')

    top_searches_block = TopSearchesBlock(
        By.CSS_SELECTOR,
        '[data-qaid="top-searches"]'
    )
    header_block = HeaderBlock(
        By.CSS_SELECTOR,
        '[data-qaid="header"]'
    )
    header_toolbar_block = HeaderToolbarBlock(
        By.CSS_SELECTOR,
        '[data-qaid="header-toolbar"]'
    )
    empty_order_popup = EmptyOrderPopup(
        By.CSS_SELECTOR,
        '[data-qaid="overlay"]'
    )
    header_reg_popup = Block(
        By.CSS_SELECTOR,
        '[data-qaid="header_registration_popup"]'
    )




from typing import NamedTuple


DEFAULT_USER_AGENT_ANDROID = (
    "Mozilla/5.0 (Linux; Android 9; Redmi Note 7) "
    "AppleWebKit/537.36 (KHTML, like Gecko) "
    "Chrome/83.0.4103.106 Mobile Safari/537.36"
)
DEFAULT_USER_AGENT_IOS = (
    "Mozilla/5.0 (iPhone; CPU iPhone OS 12_2 like Mac OS X) "
    "AppleWebKit/605.1.15 (KHTML, like Gecko) Mobile/15E148"
)
DEFAULT_USER_AGENT_HUAWEI = (
    "Mozilla/5.0 (Linux; Android 8.1; EML-L29 Build/HUAWEIEML-L29; xx-xx) "
    "AppleWebKit/537.36 (KHTML, like Gecko) Version/4.0 Chrome/65.0.3325.109 "
    "Mobile Safari/537.36"
)


class DesktopDevice(NamedTuple):
    width: int
    height: int
    user_agent: str = ''


class MobileDevice(NamedTuple):
    """
    Mobile device only can be chosen from list of known devices is generated
    from those found in the DevTools Emulation panel in Chrome Browser

    For example: "Nexus 5", "Galaxy S5", "iPhone X"

    More info: https://chromedriver.chromium.org/mobile-emulation

    Usage:
        NEXUS_5 = MobileDevice("Nexus 5")
    """
    device_name: str


class CustomDevice(NamedTuple):
    """
    It is possible to create your own device,
    which is not in the list of known devices in DevTools Emulator panel

    Usage:
        XIAOMI_REDMI_7 = CustomDevice(393, 851)
    """
    width: int
    height: int
    pixel_ratio: float = 1.0
    user_agent: str = DEFAULT_USER_AGENT_ANDROID


CHROME_DESKTOP_1920_1080 = DesktopDevice(1920, 1080)
CHROME_DESKTOP_1024_768 = DesktopDevice(1024, 768)

ANDROID_360_640 = CustomDevice(360, 640, 3.0)
ANDROID_1280_800 = CustomDevice(1280, 800, 3.0)
XIAOMI_REDMI_7 = CustomDevice(393, 851)

HUAWEI_P20 = CustomDevice(768, 850, 3.0, DEFAULT_USER_AGENT_HUAWEI)

IOS_750_1334 = CustomDevice(750, 1334, 2.0, DEFAULT_USER_AGENT_IOS)
IOS_1080_1920 = CustomDevice(1080, 1920, 2.0, DEFAULT_USER_AGENT_IOS)

NEXUS_5 = MobileDevice("Nexus 5")

IPHONE_X = MobileDevice("iPhone X")
IPAD_MINI = MobileDevice("iPad Mini")
IPAD_PRO = MobileDevice("iPad Pro")



import logging
import requests
import time

from selenium.webdriver.support.ui import WebDriverWait
from selenium.webdriver.support import expected_conditions as EC
from selenium.common.exceptions import (
    WebDriverException,
    TimeoutException
)

from promium.exceptions import PromiumException, PromiumTimeout
from promium.logger import find_console_browser_errors, repr_console_errors


log = logging.getLogger(__name__)

NORMAL_LOAD_TIME = 10


def enable_jquery(driver):
    """Enables jquery"""
    driver.execute_script(
        """
        jqueryUrl =
        'https://ajax.googleapis.com/ajax/libs/jquery/3.2.1/jquery.min.js';
        if (typeof jQuery == 'undefined') {
            var script = document.createElement('script');
            var head = document.getElementsByTagName('head')[0];
            var done = false;
            script.onload = script.onreadystatechange = (function() {
                if (!done && (!this.readyState || this.readyState == 'loaded'
                        || this.readyState == 'complete')) {
                    done = true;
                    script.onload = script.onreadystatechange = null;
                    head.removeChild(script);

                }
            });
            script.src = jqueryUrl;
            head.appendChild(script);
            console.log("injection Complete by script enable jquery")
        };
        """
    )
    wait = WebDriverWait(driver, timeout=NORMAL_LOAD_TIME)
    wait.until(
        lambda driver: (driver.execute_script(
            'return window.jQuery != undefined'
        ) is True),
        message="Jquery script injection - failed "
    )


def _check_browser_console(driver):
    """
    for using when exception occurs
    and default check console methods don`t work
    """
    console_errors = find_console_browser_errors(driver)
    if console_errors:
        log.warning(repr_console_errors(console_errors))


def wait(
        driver,
        seconds=NORMAL_LOAD_TIME,
        poll_frequency=.5,
        ignored_exceptions=None
):
    """Waits in seconds"""
    return WebDriverWait(
        driver=driver,
        timeout=seconds,
        poll_frequency=poll_frequency,
        ignored_exceptions=ignored_exceptions or [WebDriverException]
    )


def wait_until(driver, expression, seconds=NORMAL_LOAD_TIME, msg=None):
    """Waits until expression execution"""
    try:
        return wait(driver, seconds).until(expression, msg)
    except TimeoutException as e:
        _check_browser_console(driver)
        raise PromiumTimeout(e.msg, seconds, e.screen)


def wait_until_not(driver, expression, seconds=NORMAL_LOAD_TIME, msg=None):
    """Waits until not expression execution"""
    try:
        return wait(driver, seconds).until_not(expression, msg)
    except TimeoutException as e:
        _check_browser_console(driver)
        raise PromiumTimeout(e.msg, seconds, e.screen)


def wait_until_with_reload(driver, expression, trying=5, seconds=2, msg=''):
    """Waits until expression execution with page refreshing"""
    for t in range(trying):
        try:
            driver.refresh()
            if wait_until(driver, expression, seconds=seconds):
                return
        except PromiumTimeout:
            log.warning(f"Waiting `until` attempt number {t + 1}.")
    msg = f'\n{msg}' if msg else ""
    raise PromiumTimeout(
        f'The values in expression not return true after {trying} tries {msg}',
        seconds=trying * seconds,
    )


def wait_until_not_with_reload(
        driver, expression, trying=5, seconds=2, msg=''
):
    """Waits until not expression execution with page refreshing"""
    for t in range(trying):
        try:
            driver.refresh()
            if not wait_until_not(driver, expression, seconds=seconds):
                return
        except PromiumTimeout:
            log.warning(f"Waiting `until not` attempt number {t + 1}.")
    msg = f'\n{msg}' if msg else ""
    raise PromiumTimeout(
        f'The values in expression not return false after {trying} '
        f'tries {msg}.',
        seconds=trying * seconds,
    )


def wait_for_animation(driver):
    """Waits for execution animation"""
    jquery_script = 'return jQuery(":animated").length == 0;'
    return wait_until(
        driver=driver,
        expression=lambda d: d.execute_script(jquery_script),
        seconds=NORMAL_LOAD_TIME,
        msg='Animation timeout (waiting time: %s sec)' % NORMAL_LOAD_TIME
    )


def wait_document_status(driver, seconds=NORMAL_LOAD_TIME):
    wait = WebDriverWait(driver, timeout=seconds)
    try:
        wait.until(
            lambda driver: driver.execute_script(
                'return document.readyState'
            ) == 'complete',
            message=(
                "Page don't loaded because of document.readyState did not go "
                "into status 'complete'"
            )
        )
    except TimeoutException as e:
        _check_browser_console(driver)
        raise PromiumTimeout(e.msg, seconds, e.screen)


def wait_websocket_connection_open(driver, seconds=NORMAL_LOAD_TIME):
    wait = WebDriverWait(driver, timeout=seconds)
    try:
        wait.until(
            lambda driver:
            driver.execute_script('return WebSocket.OPEN') == 1,
            message="ReadyState of WebSocket connection didn't change on "
                    "OPEN"
        )
    except TimeoutException as e:
        _check_browser_console(driver)
        raise PromiumTimeout(e.msg, seconds, e.screen)


def wait_jquery_status(driver, seconds=NORMAL_LOAD_TIME):
    wait = WebDriverWait(driver, timeout=seconds)
    try:
        wait.until(
            lambda driver: driver.execute_script(
                'return jQuery.active'
            ) == 0,
            message=(
                "Page don't loaded because of jquery script did not go into "
                "'active == 0' status"
            )
        )
    except TimeoutException as e:
        _check_browser_console(driver)
        raise PromiumTimeout(e.msg, seconds, e.screen)


def wait_for_page_loaded(driver, seconds=NORMAL_LOAD_TIME):
    """Waits for page loaded"""
    wait_document_status(driver, seconds)
    enable_jquery(driver)
    wait_jquery_status(driver, seconds)
    alert_is_present = EC.alert_is_present()
    if alert_is_present(driver):
        alert_text = driver.switch_to_alert().text
        driver.switch_to_alert().dismiss()
        raise alert_text


def wait_for_alert(driver, seconds=NORMAL_LOAD_TIME):
    """Wait fo alert"""
    return WebDriverWait(
        driver, seconds, ignored_exceptions=[WebDriverException]
    ).until(
        EC.alert_is_present(), "Alert is not present."
    )


def wait_for_alert_is_displayed(driver, seconds=NORMAL_LOAD_TIME):
    try:
        wait_for_alert(driver, seconds=seconds)
    except TimeoutException:
        return False
    return True


def wait_for_status_code(url, status_code=200, tries=10, allow_redirects=True):
    """Waits for status code"""
    current_status = None
    for _ in range(tries):
        response = requests.get(
            url, allow_redirects=allow_redirects, verify=False
        )
        current_status = response.status_code
        if current_status == status_code:
            return response
        time.sleep(1)
    raise PromiumException(
        f"Url {url} return status code {current_status} after {tries} tries."
    )


def wait_until_new_window_is_opened(driver, original_window, window_num=0):
    wait_until(
        driver,
        lambda x: len(driver.window_handles) >= 2,
        msg='New window was not opened.'
    )
    window_handles = driver.window_handles
    window_handles.remove(original_window)
    return window_handles[window_num]


def wait_part_appear_in_class(driver, obj, part_class, msg=None):
    if not msg:
        msg = f'"{part_class}" not present in attribute *class*.'
    wait_until(
        driver,
        lambda driver: (part_class in obj.get_attribute('class')),
        seconds=5,
        msg=msg
    )


def wait_part_disappear_in_class(driver, obj, part_class, msg=None):
    if not msg:
        msg = f'"{part_class}" not present in attribute *class*.'
    wait_until(
        driver,
        lambda driver: (part_class not in obj.get_attribute('class')),
        seconds=5,
        msg=msg
    )


def wait_url_contains(driver, text, msg=None, timeout=5):
    if not msg:
        msg = f'"{text}" not present in current url {driver.current_url}.'
    wait_until(driver, EC.url_contains(text), seconds=timeout, msg=msg)


def wait_url_not_contains(driver, text, msg=None, timeout=5):
    if not msg:
        msg = f'"{text}" not disappear in url.'
    wait_until_not(driver, EC.url_contains(text), seconds=timeout, msg=msg)




import os

from selenium.webdriver import DesiredCapabilities, FirefoxProfile
from selenium.webdriver.chrome.options import Options as ChromeOptions
from selenium.webdriver.edge.options import Options as EdgeOptions
from selenium.webdriver.ie.options import Options as IeOptions
from selenium.webdriver.firefox.options import Options as FirefoxOptions

from promium.device_config import CustomDevice, CHROME_DESKTOP_1024_768


BASE_CAPABILITIES = {
    'loggingPrefs': {
        'performance': 'ALL',
        'server': 'ALL',
        'client': 'ALL',
        'driver': 'ALL',
        'browser': 'ALL',
    }
}


def get_download_path():
    return os.environ.get('DOWNLOAD_PATH', '/tmp')


class SettingsDrivers(object):

    @staticmethod
    def chrome():
        return SettingsChrome()

    @staticmethod
    def firefox():
        return SettingsFirefox()

    @staticmethod
    def opera():
        return SettingsOpera()

    @staticmethod
    def ie():
        return SettingsIE()

    @staticmethod
    def edge():
        return SettingsEDGE()

    @classmethod
    def get(cls, browser_name):
        name = browser_name.lower()
        if hasattr(cls, name):
            method = getattr(cls, name)
            return method()
        raise ValueError('Not correct browser name {}'.format(name))


class Base(object):
    driver_capabilities = {}
    binary_path = ''
    device = CHROME_DESKTOP_1024_768
    default_options = []
    user_options = []
    default_exp_options = {}
    user_exp_options = {}
    default_capabilities = BASE_CAPABILITIES.copy()
    user_capabilities = {}
    default_preferences = {}
    user_preferences = {}

    @classmethod
    def driver_options(cls):
        raise NotImplementedError

    @classmethod
    def options(cls):
        return cls.user_options or cls.default_options

    @classmethod
    def experimental_options(cls):
        return cls.user_exp_options or cls.default_exp_options

    @classmethod
    def capabilities(cls):
        return cls.user_capabilities or cls.default_capabilities

    @classmethod
    def preference(cls):
        return cls.user_preferences or cls.default_preferences

    @classmethod
    def set_custom_device(cls, width='', height='', pixel_ratio=1.0,
                          user_agent=''):
        """
        Set browser windows size and user agent
        :param int width: for example 1024
        :param int height: for example 768
        :param str device_name: for example Nexus 5
        :param str user_agent: not required param
        :return: namedtuple
        """
        cls.device = CustomDevice(
            width=width,
            height=height,
            pixel_ratio=pixel_ratio,
            user_agent=user_agent,
        )
        return cls.device

    @classmethod
    def set_options(cls, *args):
        """
        set browser option, not use default
        :param args: fot example --no-sandbox, --test-type, ...,
        :return: list[str]
        """
        cls.user_options.extend(args)
        return cls.user_options

    @classmethod
    def add_option(cls, option):
        """
        add browser options to default
        :param str option: for example --headless
        :return: list[str]
        """
        cls.default_options.append(option)
        return cls.default_options

    @classmethod
    def delete_option(cls, option):
        """
        deleted browser options from default
        :param str option: for example --verbose
        :return: list[str]
        """
        cls.default_options.remove(option)
        return cls.default_options

    @classmethod
    def set_windows_size(cls, device=None):
        if device:
            cls.set_custom_device(*device)

        if cls.device.user_agent:
            cls.add_option(
                "--user-agent={}".format(cls.device.user_agent)
            )

        cls.add_option(
            "--window-size={},{}".format(cls.device.width, cls.device.height)
        )
        return cls.default_options

    @classmethod
    def get_options(cls):
        cls.set_windows_size()
        driver_options = cls.driver_options()
        for option in cls.options():
            driver_options.add_argument(option)

        for exp_key, exp_option in cls.experimental_options().items():
            driver_options.add_experimental_option(exp_key, exp_option)

        if os.getenv('HEADLESS') == 'Enabled':
            driver_options.set_headless(True)

        return driver_options

    @classmethod
    def get_capabilities(cls):
        cap = cls.driver_capabilities.copy()
        for key, value in cls.capabilities().items():
            cap[key] = value
        return cap

    @classmethod
    def set_preferences(cls, **kwargs):
        cls.user_preferences.update(kwargs)
        return cls.user_preferences

    @classmethod
    def update_preferences(cls, **kwargs):
        for key, value in kwargs.items():
            cls.default_preferences[key] = value
        return cls.default_preferences

    @classmethod
    def delete_preference(cls, key):
        cls.default_preferences.pop(key)
        return cls.default_preferences

    @classmethod
    def get_preferences(cls):
        return None

    def __str__(self):
        return self.__class__.__name__


class SettingsChrome(Base):
    driver_capabilities = DesiredCapabilities.CHROME
    binary_path = '/usr/bin/chromedriver'
    default_options = [
        '--no-sandbox',
        '--disable-dev-shm-usage',
        '--no-default-browser-check',
        '--dns-prefetch-disable',
        '--no-first-run',
        '--verbose',
        '--enable-logging --v=1',
        '--test-type',
        '--ignore-certificate-errors',
        '--disable-notifications',
        '--disable-gpu'
    ]
    default_exp_options = {
        'prefs': {
            'download.default_directory': get_download_path(),
            'download.directory_upgrade': True,
            'prompt_for_download': False
        }
    }

    @classmethod
    def driver_options(cls):
        return ChromeOptions()


class SettingsOpera(SettingsChrome):
    driver_capabilities = DesiredCapabilities.OPERA
    # TODO need fix
    # binary_path = '/usr/bin/operadriver'
    binary_path = '/usr/bin/opera'

    @classmethod
    def get_options(cls):
        cls.set_windows_size()
        driver_options = cls.driver_options()
        for option in cls.options():
            driver_options.add_argument(option)

        for exp_key, exp_option in cls.experimental_options().items():
            driver_options.add_experimental_option(exp_key, exp_option)

        if os.getenv('HEADLESS') == 'Enabled':
            driver_options.set_headless(True)

        driver_options._binary_location = cls.binary_path
        return driver_options


class SettingsFirefox(Base):
    driver_capabilities = DesiredCapabilities.FIREFOX
    binary_path = '/usr/bin/gekodriver'
    default_options = ['-no-remote', ]
    default_preferences = {
        "browser.startup.homepage": "about:blank",
        "browser.download.folderList": 2,
        "browser.download.manager.showWhenStarting": False,
        "browser.download.dir": get_download_path(),
        "browser.helperApps.neverAsk.saveToDisk": "application/zip",
        "startup.homepage_welcome_url": "about:blank",
        "startup.homepage_welcome_url.additional": "about:blank",
        "pdfjs.disabled": True,
    }

    @classmethod
    def driver_options(cls):
        return FirefoxOptions()

    @classmethod
    def set_windows_size(cls, device=None):
        if device:
            cls.set_custom_device(*device)

        cls.add_option('-width {}'.format(cls.device.width))
        cls.add_option('-height {}'.format(cls.device.height))
        return cls.default_options

    @classmethod
    def get_options(cls):
        cls.set_windows_size()
        driver_options = cls.driver_options()
        for option in cls.options():
            driver_options.add_argument(option)

        if os.getenv('HEADLESS') == 'Enabled':
            driver_options.set_headless(True)

        return driver_options

    @classmethod
    def get_preferences(cls):
        profile = FirefoxProfile()
        for key, value in cls.preference().items():
            profile.set_preference(key=key, value=value)

        if cls.device.user_agent:
            profile.set_preference(
                key='general.useragent.override',
                value=cls.device.user_agent
            )
        profile.update_preferences()
        return profile


class SettingsIE(Base):
    driver_capabilities = DesiredCapabilities.INTERNETEXPLORER
    binary_path = '/usr/bin/iedriver'

    @classmethod
    def driver_options(cls):
        return IeOptions()

    @classmethod
    def get_options(cls):
        driver_options = cls.driver_options()
        for option in cls.options():
            driver_options.add_argument(option)

        return driver_options


class SettingsEDGE(Base):
    driver_capabilities = DesiredCapabilities.EDGE
    binary_path = '/usr/bin/edgedriver'

    @classmethod
    def driver_options(cls):
        return EdgeOptions()

    @classmethod
    def get_options(cls):
        return cls.driver_options()




import os

import pytest
from selenium.webdriver.firefox.options import Options as FirefoxOptions

from promium.device_config import CustomDevice
from promium.support.settings import SettingsFirefox


@pytest.mark.unit
def test_driver_capabilities():
    assert SettingsFirefox.driver_capabilities
    assert isinstance(SettingsFirefox.driver_capabilities, dict)
    assert SettingsFirefox.driver_capabilities.get('browserName') == 'firefox'


@pytest.mark.unit
def test_binary_path():
    assert SettingsFirefox.binary_path
    assert 'gekodriver' in SettingsFirefox.binary_path


@pytest.mark.unit
def test_default_options():
    assert len(SettingsFirefox.default_options) > 0
    for opt in SettingsFirefox.default_options:
        assert isinstance(opt, str)


@pytest.mark.unit
def test_default_preferences():
    assert len(SettingsFirefox.default_preferences) > 0
    for opt in SettingsFirefox.default_options:
        assert isinstance(opt, str)


@pytest.mark.unit
def test_driver_options():
    assert isinstance(SettingsFirefox.driver_options(), FirefoxOptions)


@pytest.mark.unit
def test_set_windows_size():
    SettingsFirefox.device = CustomDevice(
         width=123, height=321, user_agent='xxx'
    )
    SettingsFirefox.set_windows_size()
    ex_options = [
        '-width {}'.format(SettingsFirefox.device.width),
        '-height {}'.format(SettingsFirefox.device.height)
    ]
    for opt in ex_options:
        assert opt in SettingsFirefox.default_options


@pytest.mark.unit
def test_set_windows_size_with_device():
    device = CustomDevice(
        width=123, height=321, user_agent='user xxx'
    )
    SettingsFirefox.set_windows_size(device)
    ex_options = [
        '-width {}'.format(SettingsFirefox.device.width),
        '-height {}'.format(SettingsFirefox.device.height)
    ]
    for opt in ex_options:
        assert opt in SettingsFirefox.default_options


@pytest.mark.unit
def test_get_options_with_default_values():
    os.environ['HEADLESS'] = ''
    SettingsFirefox.user_exp_options = {}
    SettingsFirefox.user_options = []
    opt = SettingsFirefox.get_options()
    assert isinstance(opt, FirefoxOptions)
    assert SettingsFirefox.default_options == opt.arguments


@pytest.mark.unit
def test_get_options_with_user_values():
    os.environ['HEADLESS'] = ''
    SettingsFirefox.user_exp_options = {'key1': 1}
    SettingsFirefox.user_options = ['param1', 'param2', '--param3', '-p4']
    opt = SettingsFirefox.get_options()
    assert isinstance(opt, FirefoxOptions)
    assert SettingsFirefox.user_options == opt.arguments


@pytest.mark.unit
def test_get_options_without_headless():
    os.environ['HEADLESS'] = 'false'
    SettingsFirefox.user_exp_options = {}
    SettingsFirefox.user_options = []
    opt = SettingsFirefox.get_options()
    assert isinstance(opt, FirefoxOptions)
    assert '-headless' not in opt.arguments


@pytest.mark.unit
def test_get_options_with_headless():
    os.environ['HEADLESS'] = 'Enabled'
    SettingsFirefox.user_exp_options = {}
    SettingsFirefox.user_options = []
    opt = SettingsFirefox.get_options()
    assert isinstance(opt, FirefoxOptions)
    assert '-headless' in opt.arguments


@pytest.mark.skip('Fix it')
@pytest.mark.unit
def test_get_preferences():
    profile = SettingsFirefox.get_preferences()
    for pr in SettingsFirefox.default_preferences.keys():
        assert pr in profile.default_preferences


@pytest.mark.unit
def test_get_preferences_with_user_values():
    SettingsFirefox.user_preferences = {'key1': 1}
    profile = SettingsFirefox.get_preferences()
    for user_pr in SettingsFirefox.user_preferences.keys():
        assert user_pr in profile.default_preferences


@pytest.mark.unit
@pytest.mark.parametrize('user_agent', ['', None])
def test_get_preferences_with_user_agent_preferences(user_agent):
    SettingsFirefox.device = CustomDevice(
        width=123, height=321, user_agent=user_agent
    )
    profile = SettingsFirefox.get_preferences()
    assert not profile.default_preferences.get('general.useragent.override')


@pytest.mark.unit
def test_get_preferences_with_user_agent():
    SettingsFirefox.device = CustomDevice(
        width=123, height=321, user_agent='user agent xxx'
    )
    profile = SettingsFirefox.get_preferences()
    user_agent = profile.default_preferences.get('general.useragent.override')
    assert user_agent == 'user agent xxx'



import re
import json
import requests

from deepdiff import DeepDiff
from typing import Dict, Optional
from json_checker import Checker, CheckerError

from promium.common import upload_screenshot


CoerceDict = Optional[Dict[str, type]]


def base_msg(msg):
    return f"{msg}\n" if msg else ""


def _check_namedtuple(obj):
    if hasattr(obj, "_asdict"):
        return obj._asdict()
    return obj


def convert_container(container):
    if not isinstance(container, str):
        return json.dumps(
            obj=container,
            indent=4,
            sort_keys=True,
            ensure_ascii=False,
            default=lambda obj: str(obj)
        )
    return _check_namedtuple(container)


def get_text_with_ignore_whitespace_symbols(text):
    """Return text excluding spaces and whitespace_symbols"""
    text_without_whitespace_symbols = (
        text
        .replace('\t', ' ')
        .replace('\v', ' ')
        .replace('\r', ' ')
        .replace('\n', ' ')
        .replace('\f', ' ')
        .strip()
    )
    text_list = text_without_whitespace_symbols.split(' ')
    text_list_without_space = [word for word in text_list if word]
    needful_text = ' '.join(text_list_without_space)
    return needful_text


class BaseSoftAssertion(object):

    # TODO is not cleaned in unit tests need use __init__
    assertion_errors = []

    def soft_assert_true(self, expr, msg=None):
        """Check that the expression is true."""
        if not expr:
            error = "Is not true." if not msg else msg
            self.assertion_errors.append(f"{error}\n")
            return error

    def soft_assert_false(self, expr, msg=None):
        """Check that the expression is false."""
        if expr:
            message = "Is not false." if not msg else msg
            self.assertion_errors.append(f"{message}\n")
            return message

    def soft_assert_equals(self, current, expected, msg=None, show_diff=False):
        """Just like self.soft_assert_true(current == expected)"""
        message = (
            f"Current and expected has different data types: "
            f"current is {type(current)}, expected is {type(expected)}\n"
            if type(current) != type(expected) else ""
        )
        difference = ""
        diff = None
        if type(current) is list and type(expected) is list:
            diff = DeepDiff(current, expected, ignore_order=True)
        if show_diff:
            if type(current) is str and type(expected) is str:
                current_splitlines = current.splitlines()
                expected_splitlines = expected.splitlines()
                if len(current_splitlines) > 1 or len(expected_splitlines) > 1:
                    diff = DeepDiff(current_splitlines, expected_splitlines)
        if diff:
            difference = f"\nDifference:\n{convert_container(diff)}"
        assert_message = (
            f"{base_msg(msg)}\n"
            f"Current - {convert_container(current)}\n"
            f"Expected - {convert_container(expected)}\n"
            f"{difference}"
            f"{message}"
        )
        return self.soft_assert_true(current == expected, assert_message)

    def soft_assert_not_equals(self, current, expected, msg=None):
        """Just like self.soft_assert_true(current != expected)"""
        message = (
            f"{base_msg(msg)}"
            f"Current - {convert_container(current)}\n"
            f"Expected - {convert_container(expected)}\n"
        )
        self.soft_assert_false(current == expected, message)

    def soft_assert_in(self, member, container, msg=None):
        """Just like self.soft_assert_true(member IN container)"""
        msg = (
            f"{base_msg(msg)}"
            f"{member} not found in {convert_container(container)}\n"
        )
        return self.soft_assert_true(member in container, msg)

    def soft_assert_not_in(self, member, container, msg=None):
        """Just like self.soft_assert_true(member NOT IN container)"""
        msg = (
            f"{base_msg(msg)}"
            f"{member} unexpectedly found in {convert_container(container)}\n"
        )
        return self.soft_assert_true(member not in container, msg)

    def soft_assert_less_equal(self, a, b, msg=None):
        """Just like self.soft_assert_true(a <= b)"""
        error = f"{base_msg(msg)}{a} not less than or equal to {b}\n"
        return self.soft_assert_true(a <= b, error)

    def soft_assert_less(self, a, b, msg=None):
        """Just like self.soft_assert_true(a < b)"""
        error = f"{base_msg(msg)}{a} not less than {b}\n"
        return self.soft_assert_true(a < b, error)

    def soft_assert_greater_equal(self, a, b, msg=None):
        """Just like self.soft_assert_true(a >= b)"""
        error = f"{base_msg(msg)}{a} not greater than or equal to {b}\n"
        return self.soft_assert_true(a >= b, error)

    def soft_assert_greater(self, a, b, msg=None):
        """Just like self.soft_assert_true(a > b)"""
        error = f"{base_msg(msg)}{a} not greater than {b}\n"
        return self.soft_assert_true(a > b, error)

    def soft_assert_regexp_matches(self, text, expected_regexp, msg=None):
        """Fail the test unless the text matches the regular expression."""
        pattern = re.compile(expected_regexp)
        result = pattern.search(text)

        if not result:
            error = (
                f"{base_msg(msg)}"
                f"Regexp didn't match."
                f"Pattern {str(pattern.pattern)} not found in {str(text)}\n"
            )
            self.assertion_errors.append(error)
            return error

    def soft_assert_disable(self, element, msg=None):
        """Check that the obj hasn't attribute."""
        default_msg = f"Not disabled {element}\n" if not msg else ""
        error = f"{base_msg(msg)}{default_msg}"
        return self.soft_assert_true(element.get_attribute("disabled"), error)

    def soft_assert_is_none(self, obj, msg=None):
        """Same as self.soft_assert_true(obj is None)."""
        default_msg = f"{obj} is not None.\n" if not msg else ""
        error = f"{base_msg(msg)}{default_msg}"
        return self.soft_assert_true(obj is None, error)

    def soft_assert_is_not_none(self, obj, msg=None):
        """Included for symmetry with self.soft_assert_is_none."""
        default_msg = "Unexpectedly None.\n" if not msg else ""
        error = f"{base_msg(msg)}{default_msg}"
        return self.soft_assert_true(obj is not None, error)

    def soft_assert_is_instance(self, obj, cls, msg=None):
        """Same as self.soft_assert_true(isinstance(obj, cls))"""
        default_msg = (
            f"{obj} is not an instance of {cls}.\n" if not msg else ""
        )
        error = f"{base_msg(msg)}{default_msg}"
        return self.soft_assert_true(isinstance(obj, cls), error)

    def soft_assert_equals_text_with_ignore_spaces_and_register(
        self,
        current_text,
        expected_text,
        msg='Invalid checked text.'
    ):
        """Checking of text excluding spaces and register"""
        current = get_text_with_ignore_whitespace_symbols(current_text)
        expected = get_text_with_ignore_whitespace_symbols(expected_text)
        if not current:
            msg = "Warning: current text is None!"
        self.soft_assert_equals(
            current.lower(),
            expected.lower(),
            f"{msg}\nCurrent text without formating: {current_text}"
            f"\nExpected text without formating: {expected_text}"
        )

    def soft_assert_schemas(self, current, expected, msg=''):
        """
        Example:
            {'test1': 1} == {'test1: int}

        :param dict current: current response
        :param dict expected: expected dict(key: type)
        :param str msg:
        :return error
        """
        try:
            Checker(expected, soft=True).validate(current)
        except CheckerError as e:
            error = f'{msg}\n{e}\n'
            self.assertion_errors.append(error)
            return error

    def assert_keys_and_instances(
        self,
        actual_dict,
        expected_dict,
        can_be_null=None,
        msg=None
    ):
        """
        :param dict actual_dict:
        :param dict expected_dict:
        :param list | None can_be_null: must be if default value None
        :param basestring msg:
        """
        assert actual_dict, 'Actual dict is empty, check your data'

        self.soft_assert_equals(
            sorted(iter(actual_dict.keys())),
            sorted(iter(expected_dict.keys())),
            'Wrong keys list.'
        )
        for actual_key, actual_value in actual_dict.items():
            self.soft_assert_in(
                member=actual_key,
                container=expected_dict,
                msg=f'Not expected key "{actual_key}".'
            )
            if actual_key in expected_dict:
                expected_value = (
                    type(None) if
                    actual_value is None and
                    actual_key in (can_be_null or []) else
                    expected_dict[actual_key]
                )
                message = f'({msg})' if msg else ''
                self.soft_assert_true(
                    expr=isinstance(actual_value, expected_value),
                    msg=(
                        f'Wrong object instance class.\n'
                        f'Key "{actual_key}" value is "{type(actual_value)}", '
                        f'expected "{expected_value}". {message}'
                    )
                )

    def soft_assert_dicts_with_ignore_types(self, current, expected, msg=''):
        expected_dict = _check_namedtuple(expected)
        current_dict = _check_namedtuple(current)
        errors = ''
        for actual_key, actual_value in current_dict.items():
            if actual_key in expected_dict:
                if str(actual_value) != str(expected_dict[actual_key]):
                    errors += (
                        f"\nKey {actual_key} "
                        f"\nHas not correct value:  {actual_value}  "
                        f"Must be have this:  {expected_dict[actual_key]} \n"
                    )
        if errors:
            msg += (
                f'{errors}'
                f"\nExpected dict: {convert_container(expected_dict)}"
                f"\nCurrent dict: {convert_container(current_dict)}"
            )
            self.assertion_errors.append(msg)

    def soft_assert_equal_dict_from_clerk_magic(
        self,
        current_dict,
        expected_dict,
        msg=None,
        coerce: CoerceDict = None
    ):
        """
        Comparison by type, if the types do not match - compares in a string
        If the dict key was lost, print him in missing list.
        """
        def _equal(
                field_name: str,
                current: str, expected: str,
                type_coerce: CoerceDict
        ):
            if type_coerce is not None and field_name in type_coerce:
                cast = type_coerce[field_name]
                return cast(current) == cast(expected)
            elif type(current) == type(expected):
                return current == expected
            else:
                return str(current) == str(expected)

        missing = []
        unequal = []
        for key, value in expected_dict.items():
            if key not in current_dict:
                missing.append((key, value))
            elif not _equal(key, value, current_dict[key], coerce):
                unequal.append((
                    key, f"\n\t\tproduct_data - '{value}'"
                         f" != analytic_data - '{current_dict[key]}'"
                ))
        self.soft_assert_true(
            (missing, unequal) == ([], []),
            msg=(
                "{base_msg}\n"
                "Unequal_keys:\n\t{unequal}\n"
                "\nMissed_keys:\n\t[{missing}]".format(
                    base_msg=base_msg(msg),
                    unequal='\n\t'.join(
                        [":".join(map(str, a)) for a in unequal]
                    ),
                    missing='\n\t'.join(
                        [":".join(map(str, a)) for a in missing]
                    ),
                )
            )
        )


class RequestSoftAssertion(BaseSoftAssertion):

    @property
    def url(self):
        return self.session.url

    def base_msg(self, msg=None):
        url = f"\n{self.url}" if self.url else ""
        exception = f"\n{msg}" if msg else ""
        return f"{url}{exception}"


class WebDriverSoftAssertion(BaseSoftAssertion):

    @property
    def driver(self):
        return self.driver

    @property
    def url(self):
        return self.driver.current_url

    @property
    def assertion_screenshot(self):
        return upload_screenshot(self.driver)

    def base_msg(self, default_msg=None, extend_msg=None):
        default = f"Default message: {default_msg}\n" if default_msg else ""
        extend = f"Extend message: {extend_msg}\n" if extend_msg else ""
        return (
            f"{default}{extend}"
            f"URL: {self.url}\n"
            f"Screenshot: {self.assertion_screenshot}\n"
        )

    def soft_assert_page_title(self, expected_title, msg=None):
        default_msg = 'Wrong page title.'
        error = f"{self.base_msg(default_msg=default_msg, extend_msg=msg)}"
        self.soft_assert_equals(self.driver.title, expected_title, error)

    def soft_assert_current_url(self, expected_url, msg=None):
        default_msg = 'Wrong current url.'
        error = f"{self.base_msg(default_msg=default_msg, extend_msg=msg)}"
        self.soft_assert_equals(self.url, expected_url, error)

    def soft_assert_current_url_contains(self, url_mask, msg=None):
        default_msg = f"URL {str(self.url)} doesn't contains {str(url_mask)}."
        error = f"{self.base_msg(default_msg=default_msg, extend_msg=msg)}"
        self.soft_assert_in(url_mask, self.url, error)

    def soft_assert_current_url_not_contains(self, url_mask, msg=None):
        default_msg = f"URL {str(self.url)} contains {str(url_mask)}."
        error = f"{self.base_msg(default_msg=default_msg, extend_msg=msg)}"
        self.soft_assert_not_in(url_mask, self.url, error)

    def soft_assert_element_is_present(self, element, msg=None):
        default_msg = (
            f"Element {element.by}={element.locator} is not present "
            f"on page at current time."
        )
        error = f"{self.base_msg(default_msg=default_msg, extend_msg=msg)}"
        self.soft_assert_true(element.is_present(), error)

    def soft_assert_element_is_not_present(self, element, msg=None):
        default_msg = (
            f"Element {element.by}={element.locator} is found on page."
        )
        error = f"{self.base_msg(default_msg=default_msg, extend_msg=msg)}"
        self.soft_assert_false(element.is_present(), error)

    def soft_assert_element_is_displayed(self, element, msg=None):
        default_msg = (
            f"Element {element.by}={element.locator} is "
            f"not visible to a user."
        )
        error = f"{self.base_msg(default_msg=default_msg, extend_msg=msg)}"
        self.soft_assert_true(element.is_displayed(), error)

    def soft_assert_element_is_not_displayed(self, element, msg=None):
        default_msg = (
            f"Element {element.by}={element.locator} is visible to a user."
        )
        error = f"{self.base_msg(default_msg=default_msg, extend_msg=msg)}"
        self.soft_assert_false(element.is_displayed(), error)

    def soft_assert_element_displayed_in_viewport(self, element, msg=None):
        """This method checks that element is viewable in viewport"""
        element_in_viewport = self.driver.execute_script(
            """
            function elementInViewport(el) {
                var top = el.offsetTop;
                var left = el.offsetLeft;
                var width = el.offsetWidth;
                var height = el.offsetHeight;


                while(el.offsetParent) {
                el = el.offsetParent;
                top += el.offsetTop;
                left += el.offsetLeft;
                }

                return (
                    top >= window.pageYOffset &&
                    left >= window.pageXOffset &&
                    (top + height) <= (window.pageYOffset + window.innerHeight)
                    &&
                    (left + width) <= (window.pageXOffset + window.innerWidth)
                );
            }

            element = arguments[0];
            return elementInViewport(element);
            """, element.lookup()
        )

        if element_in_viewport is not True:
            default_msg = (
                f"Element {element.by}={element.locator} "
                f"not displayed in viewport"
            )
            error = f"{self.base_msg(default_msg=default_msg, extend_msg=msg)}"
            self.assertion_errors.append(error)

    def soft_assert_image_status_code(
        self, image, timeout=2, status_code=200, msg=None
    ):
        if not image.get_attribute('src'):
            return self.assertion_errors.append(
                'Image does not have attribute "src"'
            )

        img_url = image.get_attribute('src')
        response = requests.get(img_url, timeout=timeout, verify=False)

        self.soft_assert_equals(
            response.status_code,
            status_code,
            msg=msg if msg else 'img status code != 200'
        )
